import pygame
import typing
from mili import typing as _typing
from mili.mili import MILI as _MILI
from mili._utils._interaction import InteractionCursor
from mili import icon as _icon
from mili import error as _error
from mili import animation as _anim
from mili import _coreutils


class GenericApp:
    def __init__(
        self,
        window: pygame.Window,
        target_framerate: int = 60,
        clear_color: pygame.typing.ColorLike | None = 0,
        start_style: _typing.ElementStyleLike | None = None,
        use_global_mouse: bool = False,
    ):
        self.window = window
        self.clock = pygame.Clock()
        self.target_framerate: int = target_framerate
        self.mili = _MILI(self.window.get_surface(), use_global_mouse)
        self.clear_color: pygame.typing.ColorLike | None = clear_color
        self.start_style: _typing.ElementStyleLike | None = start_style
        self.delta_time: float = 0
        self.running = True

    def update(self): ...

    def ui(self): ...

    def event(self, event: pygame.Event): ...

    def quit(self):
        self.on_quit()
        pygame.quit()
        raise SystemExit

    def on_quit(self): ...

    def post_draw(self): ...

    def run(self):
        while self.running:
            self.mili.start(self.start_style, window_position=self.window.position)
            for event in pygame.event.get():
                if event.type == pygame.WINDOWCLOSE and event.window == self.window:
                    self.quit()
                    return
                else:
                    self.event(event)

            if self.clear_color is not None:
                self.window.get_surface().fill(self.clear_color)
            self.update()
            self.ui()
            self.mili.update_draw()
            self.post_draw()
            self.window.flip()
            self.delta_time = self.clock.tick(self.target_framerate) / 1000


def _titlebar_button_style(
    style, icon, bg_hover=(60, 60, 60)
) -> _typing._UIAppButtonStyleLike:
    if style is None:
        style = {}
    bg = style.get("bg_color", {})
    alpha = style.get("alpha", {})
    return {
        "bg_color": {
            "default": bg.get("default", (40, 40, 40)),
            "hover": bg.get("hover", bg_hover),
            "press": bg.get("press", (32, 32, 32)),
        },
        "icon": style.get("icon", icon),
        "alpha": {
            "default": alpha.get("default", 180),
            "hover": alpha.get("hover", 255),
            "press": alpha.get("press", 150),
        },
    }


class UIApp:
    def __init__(
        self,
        window: pygame.Window,
        style: _typing.UIAppStyleLike | dict | None = None,
        use_global_mouse: bool = False,
    ):
        pygame.init()
        self.window = window
        if style is None:
            style = {}
        self.style: _typing.UIAppStyleLike = {
            "start_style": style.get("start_style", {}),
            "target_framerate": style.get("target_framerate", 60),
            "clear_color": style.get("clear_color", (22, 22, 22)),
            "window_outline": style.get("window_outline", 1),
            "window_icon": style.get("window_icon", None),
            "outline_color": style.get("outline_color", (40, 40, 40)),
            "fullscreen_key": style.get("fullscreen_key", pygame.K_F11),
            "titlebar_show_title": style.get("titlebar_show_title", True),
            "title_style": style.get(
                "title_style", {"size": 18, "color": (180, 180, 180)}
            ),
            "titlebar_color": style.get("titlebar_color", (40, 40, 40)),
            "use_appdata_folder": style.get("use_appdata_folder", True),
            "minimize_button": _titlebar_button_style(
                style.get("minimize_button", {}),
                _icon.lazy_colored("iconify", "minimize", "material-symbols", "white"),
            ),
            "maximize_button": _titlebar_button_style(
                style.get("maximize_button", {}),
                _icon.lazy_colored(
                    "iconify", "square-rounded", "material-symbols", "white"
                ),
            ),
            "close_button": _titlebar_button_style(
                style.get("close_button", {}),
                _icon.lazy_colored("iconify", "close", "material-symbols", "white"),
                "red",
            ),
        }
        self.clock = pygame.Clock()
        self.delta_time: float = 0
        self.mili = _MILI(self.window.get_surface(), use_global_mouse)
        self.win_borders = CustomWindowBorders(self.window)
        self.win_behavior = CustomWindowBehavior(
            self.window, self.win_borders, pygame.display.get_desktop_sizes()[0]
        )
        self.adaptive_scaler = AdaptiveUIScaler(self.window, self.window.size)
        self.scale = self.adaptive_scaler.scale
        self.running = True
        InteractionCursor.setup(update_id="cursor")
        _coreutils._number_mods["s"] = self.adaptive_scaler.scale
        if self.style["use_appdata_folder"]:
            import os

            if not os.path.exists("appdata"):
                os.mkdir("appdata")
            _icon.setup("appdata", "white")

    def update(self): ...

    def ui(self): ...

    def ui_tbar_pretitle(self): ...
    def ui_tbar_posttitle(self): ...
    def ui_tbar_prebuttons(self): ...

    def event(self, event: pygame.Event): ...

    def quit(self):
        self.on_quit()
        pygame.quit()
        raise SystemExit

    def on_quit(self): ...

    def post_draw(self): ...

    def run(self):
        while self.running:
            style = self.style["start_style"]
            if self.window.borderless:
                style = {"pad": 0, "spacing": 0}
            else:
                self.win_borders.active = False
            self.mili.start(style, window_position=self.window.position)
            for event in pygame.event.get():
                if event.type == pygame.WINDOWCLOSE and event.window == self.window:
                    self.quit()
                    return
                else:
                    self.event(event)
                if event.type == pygame.KEYDOWN:
                    if event.key == self.style["fullscreen_key"]:
                        self.win_behavior.toggle_fullscreen()
                    if event.key == pygame.K_ESCAPE and self.win_behavior.fullscreen:
                        self.win_behavior.fullscreen_off()

            ccolor = self.style["clear_color"]
            self.win_borders.update()
            self.adaptive_scaler.update()
            _anim.update_all()
            self.update()
            outline, outline_col = (
                self.style["window_outline"],
                self.style["outline_color"],
            )
            if self.window.borderless and not self.win_behavior.fullscreen:
                wsize = self.window.size
                h = self.win_borders.titlebar_height
                with self.mili.begin(
                    (0, 0, wsize[0], h),
                    {"axis": "x", "pad": 0, "spacing": 0, "z": 9999},
                ):
                    self.mili.rect(
                        {"color": self.style["titlebar_color"], "border_radius": 0}
                    )
                    self.ui_tbar_pretitle()
                    if (icon := self.style["window_icon"]) is not None:
                        self.mili.image_element(icon, {"pad": 1}, (0, 0, h, h))
                    if self.style["titlebar_show_title"]:
                        self.mili.text_element(
                            self.window.title,
                            self.style["title_style"] | {"growy": False, "growx": True},
                            (0, 0, 0, h),
                        )
                    self.ui_tbar_posttitle()
                    self.mili.element(None, {"fillx": True})
                    self.ui_tbar_prebuttons()
                    for name, button in [
                        ("minimize", self.style["minimize_button"]),
                        ("maximize", self.style["maximize_button"]),
                        ("close", self.style["close_button"]),
                    ]:
                        if button is None:
                            continue
                        surf = _icon._get_icon(button["icon"])
                        if surf is None:
                            continue
                        with self.mili.element((0, 0, h, h)) as btn:
                            bg = _coreutils._get_conditional_color(
                                btn, button["bg_color"]
                            )
                            alpha = _coreutils._get_conditional_color(
                                btn, button["alpha"]
                            )
                            self.mili.rect({"color": bg, "border_radius": 0})
                            self.mili.image(surf, {"alpha": alpha, "smoothscale": True})
                            if btn.left_clicked and self.win_borders.can_interact():
                                match name:
                                    case "close":
                                        self.quit()
                                    case "maximize":
                                        self.win_behavior.toggle_maximize()
                                    case "minimize":
                                        self.win_behavior.minimize()
                with self.mili.begin(
                    (0, 0, wsize[0], wsize[1] - h), self.style["start_style"]
                ):
                    if ccolor is not None:
                        self.mili.rect({"color": ccolor, "border_radius": 0})
                    if outline:
                        self.mili.rect(
                            {
                                "color": outline_col,
                                "outline": 1,
                                "draw_above": True,
                                "border_radius": 0,
                            }
                        )
                    try:
                        self.mili.id_checkpoint(10)
                    except _error.MILIStatusError:
                        ...
                    self.ui()
            else:
                if ccolor is not None:
                    self.mili.rect({"color": ccolor, "border_radius": 0})
                if outline:
                    self.mili.rect(
                        {
                            "color": outline_col,
                            "outline": 1,
                            "draw_above": True,
                            "border_radius": 0,
                        }
                    )
                try:
                    self.mili.id_checkpoint(10)
                except _error.MILIStatusError:
                    ...
                self.ui()
            InteractionCursor.apply()
            self.mili.update_draw()
            self.post_draw()
            self.window.flip()
            self.delta_time = self.clock.tick(self.style["target_framerate"]) / 1000


class _HasSizeAttr(typing.Protocol):
    size: typing.Sequence[float]


class AdaptiveUIScaler:
    def __init__(
        self,
        window: pygame.Window | _HasSizeAttr,
        relative_size: typing.Sequence[float],
        scale_clamp: tuple[float | None, float | None] | None = (0.2, 1.2),
        width_weight: float = 1,
        height_weight: float = 1,
        min_value: float | None = 1,
        int_func: typing.Callable[[float], int] = int,
    ):
        self.window = window
        self.relative_size = pygame.Vector2(relative_size)
        self.scale_clamp = scale_clamp
        self.width_weight = width_weight
        self.height_weight = height_weight
        self.min_value = min_value
        self.int_func = int_func
        self.extra_scale: float = 1
        self._mult = 1

    def update(self):
        size = self.window.size
        xmult = size[0] / self.relative_size.x
        ymult = size[1] / self.relative_size.y
        mult = (xmult * self.width_weight + ymult * self.height_weight) / (
            self.width_weight + self.height_weight
        )
        if self.scale_clamp is not None:
            clampmin, clampmax = self.scale_clamp
            if clampmin is not None and mult < clampmin:
                mult = clampmin
            if clampmax is not None and mult > clampmax:
                mult = clampmax
        self._mult = mult

    def scale(self, value: float, min_override: float | None = None) -> int:
        min_value = self.min_value
        if min_override is not None:
            min_value = min_override
        scaled = value * self._mult * self.extra_scale
        if min_value is not None and scaled < min_value:
            scaled = min_value
        return self.int_func(scaled)

    def scalef(self, value: float, min_override: float | None = None) -> float:
        min_value = self.min_value
        if min_override is not None:
            min_value = min_override
        scaled = value * self._mult * self.extra_scale
        if min_value is not None and scaled < min_value:
            scaled = min_value
        return scaled


class CustomWindowBorders:
    def __init__(
        self,
        window: pygame.Window,
        border_size: int = 3,
        corner_size: int = 6,
        titlebar_height: int = 25,
        change_cursor: bool = True,
        minimum_ratio: float | None = None,
        ratio_axis: typing.Literal["x", "y"] = "y",
        uniform_resize_key: int | None = pygame.K_LSHIFT,
        allowed_directions: list[
            typing.Literal[
                "top",
                "left",
                "bottom",
                "right",
                "topleft",
                "topright",
                "bottomleft",
                "bottomright",
            ]
        ]
        | typing.Literal["all"] = "all",
        constraint_rect: pygame.typing.RectLike | None = None,
        *,
        on_move: typing.Callable[[], None] | None = None,
        on_resize: typing.Callable[[], None] | None = None,
        on_start_move: typing.Callable[[], None] | None = None,
        on_start_resize: typing.Callable[[], None] | None = None,
        on_end_move: typing.Callable[[], None] | None = None,
        on_end_resize: typing.Callable[[], None] | None = None,
    ):
        self.window = window
        self.border_size = border_size
        self.corner_size = corner_size
        self.titlebar_height = titlebar_height
        self.change_cursor = change_cursor
        self.minimum_ratio = minimum_ratio
        self.ratio_axis = ratio_axis
        self.uniform_resize_key = uniform_resize_key
        self.allowed_directions = allowed_directions
        self.constraint_rect = (
            pygame.Rect(constraint_rect) if constraint_rect is not None else None
        )
        self._callbacks: dict[
            typing.Literal["move", "resize"],
            dict[
                typing.Literal["start", "end", "during"],
                typing.Callable[[], None] | None,
            ],
        ] = {
            "move": {"start": on_start_move, "end": on_end_move, "during": on_move},
            "resize": {
                "start": on_start_resize,
                "end": on_end_resize,
                "during": on_resize,
            },
        }
        self.dragging: bool = False
        self.resizing: bool = False
        self.relative: pygame.Vector2 = pygame.Vector2()
        self.cumulative_relative: pygame.Vector2 = pygame.Vector2()
        self.active = True
        self.active_drag = True
        self.active_resize = True

        self._double_click_cooldown = 100
        self._toggle_maximize_func: typing.Callable | None = None
        self._left_pressed = False
        self._press_count = 0
        self._press_time = 0
        self._press_rel = pygame.Vector2()
        self._press_global = pygame.Vector2()
        self._start_val = pygame.Vector2()
        self._resize_winsize = None
        self._resize_winpos = None
        self._resize_handle: CustomWindowBorders._ResizeHandle | None = None
        self._resize_handles = [
            CustomWindowBorders._ResizeHandle(
                self, "topleft", True, None, "xy", pygame.SYSTEM_CURSOR_SIZENWSE
            ),
            CustomWindowBorders._ResizeHandle(
                self, "topright", True, None, "y", pygame.SYSTEM_CURSOR_SIZENESW
            ),
            CustomWindowBorders._ResizeHandle(
                self, "bottomleft", True, None, "x", pygame.SYSTEM_CURSOR_SIZENESW
            ),
            CustomWindowBorders._ResizeHandle(
                self, "bottomright", True, None, None, pygame.SYSTEM_CURSOR_SIZENWSE
            ),
            CustomWindowBorders._ResizeHandle(
                self, "top", False, "x", "y", pygame.SYSTEM_CURSOR_SIZENS
            ),
            CustomWindowBorders._ResizeHandle(
                self, "left", False, "y", "x", pygame.SYSTEM_CURSOR_SIZEWE
            ),
            CustomWindowBorders._ResizeHandle(
                self, "bottom", False, "x", None, pygame.SYSTEM_CURSOR_SIZENS
            ),
            CustomWindowBorders._ResizeHandle(
                self, "right", False, "y", None, pygame.SYSTEM_CURSOR_SIZEWE
            ),
        ]

    def can_interact(self):
        return self.window.focused and self.cumulative_relative.length() == 0

    def mouse_changed(self):
        self._press_rel = pygame.mouse.get_pos()
        self._press_global = pygame.Vector2(pygame.mouse.get_pos(True))
        self._start_val = pygame.Vector2(
            self.window.position if self.dragging else self.window.size
        )

    def update(self) -> bool:
        if not self.window.focused or not self.active:
            InteractionCursor._active = True
            return False

        mpos = pygame.Vector2(pygame.mouse.get_pos())
        gmpos = pygame.Vector2(pygame.mouse.get_pos(True))
        pressed = pygame.mouse.get_pressed(5, True)[0]
        just = pressed and not self._left_pressed
        self._left_pressed = pressed
        modified_cursor = False
        tbar_rect = pygame.Rect(0, 0, self.window.size[0], self.titlebar_height)

        for handle in self._resize_handles:
            if (
                self.allowed_directions == "all"
                or handle._name in self.allowed_directions
            ):
                handle._make_rect()
            else:
                handle._rect = pygame.Rect()

        if self.change_cursor:
            next_cursor = None
            if not self.resizing and self.active_drag:
                if tbar_rect.h > 0 and (tbar_rect.collidepoint(mpos) or self.dragging):
                    next_cursor = pygame.SYSTEM_CURSOR_HAND
                    modified_cursor = True
            if not self.dragging and self.active_resize:
                for handle in self._resize_handles:
                    if (
                        self.allowed_directions == "all"
                        or handle._name in self.allowed_directions
                    ) and (
                        handle._rect.collidepoint(mpos)
                        or (self.resizing and self._resize_handle is handle)
                    ):
                        pygame.mouse.set_cursor(handle._cursor)
                        modified_cursor = True
                        next_cursor = None
                        break
            if next_cursor is not None:
                pygame.mouse.set_cursor(next_cursor)

        if self.active_drag:
            if just:
                for handle in self._resize_handles:
                    if (
                        self.allowed_directions == "all"
                        or handle._name in self.allowed_directions
                    ) and handle._rect.collidepoint(mpos):
                        self.resizing = True
                        self._resize_handle = handle
                        self._press_global = gmpos
                        self._press_rel = mpos
                        self._resize_winpos = self.window.position
                        self._resize_winsize = self.window.size
                        self._start_val = pygame.Vector2(self.window.size)
                        callback = self._callbacks["resize"]["start"]
                        if callback is not None:
                            callback()
                        break

            if pressed:
                if self.resizing and self._resize_handle:
                    self._resize_handle._update(gmpos)
            else:
                if self.resizing:
                    self.cumulative_relative = self.window.size - self._start_val
                    callback = self._callbacks["resize"]["end"]
                    if callback is not None:
                        callback()
                self.resizing = False
                self._resize_handle = None
                self.relative = pygame.Vector2()
                self.cumulative_relative = pygame.Vector2()

        if self.resizing or tbar_rect.h <= 0:
            InteractionCursor._active = not modified_cursor
            return modified_cursor

        if self.active_drag:
            if just and tbar_rect.collidepoint(mpos):
                if self._press_count == 0:
                    self._press_count += 1
                    self._press_time = pygame.time.get_ticks()
                else:
                    if (
                        pygame.time.get_ticks() - self._press_time
                        <= self._double_click_cooldown
                        and self._toggle_maximize_func is not None
                    ):
                        self._toggle_maximize_func()
                    self._press_count = 0
                    self._press_time = 0
                self._press_rel = mpos
                self._press_global = gmpos
                self._start_val = pygame.Vector2(self.window.position)
                self.dragging = True
                callback = self._callbacks["move"]["start"]
                if callback is not None:
                    callback()

            if pressed:
                if not just and self.dragging:
                    previous = pygame.Vector2(self.window.position)
                    new_position = self._start_val + (gmpos - self._press_global)
                    self.window.position = self._constrain_position(new_position)
                    self.relative = self.window.position - previous
                    self.cumulative_relative = self.window.position - self._start_val
                    callback = self._callbacks["move"]["during"]
                    if callback is not None:
                        callback()
            else:
                if self.dragging:
                    self.cumulative_relative = self.window.position - self._start_val
                    callback = self._callbacks["move"]["end"]
                    if callback is not None:
                        callback()
                self.dragging = False
                self.relative = pygame.Vector2()
                self.cumulative_relative = pygame.Vector2()
        InteractionCursor._active = not modified_cursor
        return modified_cursor

    def _constrain_position(self, pos: pygame.Vector2):
        if self.constraint_rect is None:
            return pos
        sx, sy = self.window.size
        return pygame.Vector2(
            pygame.math.clamp(
                pos.x,
                self.constraint_rect.x,
                max(self.constraint_rect.right - sx, self.constraint_rect.x),
            ),
            pygame.math.clamp(
                pos.y,
                self.constraint_rect.y,
                max(self.constraint_rect.bottom - sy, self.constraint_rect.y),
            ),
        )

    class _ResizeHandle:
        def __init__(
            self,
            parent: "CustomWindowBorders",
            name,
            iscorner,
            axis,
            movewindow,
            cursor,
        ):
            self._parent = parent
            self._name = name
            self._iscorner = iscorner
            self._axis = axis
            self._movewindow = movewindow
            self._cursor = cursor
            self._axis_lock = (
                "x" if self._axis == "y" else "y" if self._axis == "x" else None
            )
            self._rect: pygame.Rect = pygame.Rect()

        def _make_rect(self):
            if self._iscorner:
                rect = pygame.Rect(
                    0, 0, self._parent.corner_size, self._parent.corner_size
                )
            else:
                rect = pygame.Rect(
                    0,
                    0,
                    self._parent.window.size[0]
                    if self._axis == "x"
                    else self._parent.border_size,
                    self._parent.window.size[1]
                    if self._axis == "y"
                    else self._parent.border_size,
                )
            if self._name == "topright":
                rect = rect.move_to(topright=(self._parent.window.size[0], 0))
            elif self._name == "bottomleft" or self._name == "bottom":
                rect = rect.move_to(bottomleft=(0, self._parent.window.size[1]))
            elif self._name == "bottomright" or self._name == "right":
                rect = rect.move_to(bottomright=self._parent.window.size)
            self._rect = rect

        def _update(self, gmpos):
            if self._parent._resize_winsize is None:
                return
            previous = pygame.Vector2(self._parent.window.size)
            rel: pygame.Vector2 = gmpos - self._parent._press_global
            posrel = pygame.Vector2()
            if self._axis_lock == "y":
                rel.x = 0
            elif self._axis_lock == "x":
                rel.y = 0
            if self._movewindow == "x":
                rel.x *= -1
            elif self._movewindow == "y":
                rel.y *= -1
            elif self._movewindow == "xy":
                rel.x *= -1
                rel.y *= -1
            if (
                self._parent.uniform_resize_key is not None
                and pygame.key.get_pressed()[self._parent.uniform_resize_key]
            ):
                if rel.x != rel.y:
                    if abs(rel.x) > abs(rel.y):
                        rel.y = rel.x
                    else:
                        rel.x = rel.y
            if self._movewindow == "x":
                posrel.x = rel.x
            elif self._movewindow == "y":
                posrel.y = rel.y
            elif self._movewindow == "xy":
                posrel = rel
            newsize = self._parent._resize_winsize + rel
            if newsize.x <= 0 or newsize.y <= 0:
                return
            if self._parent.minimum_ratio is not None:
                if self._parent.ratio_axis == "y":
                    ratio = newsize[0] / newsize[1]
                else:
                    ratio = newsize[1] / newsize[0]
                if ratio < self._parent.minimum_ratio:
                    if self._parent.ratio_axis == "y":
                        newsize = pygame.Vector2(
                            newsize[1] * self._parent.minimum_ratio, newsize[1]
                        )
                    else:
                        newsize = pygame.Vector2(
                            newsize[0], newsize[0] * self._parent.minimum_ratio
                        )
            if self._parent.constraint_rect is not None:
                newsize = pygame.Vector2(
                    pygame.math.clamp(newsize.x, 0, self._parent.constraint_rect.w),
                    pygame.math.clamp(newsize.y, 0, self._parent.constraint_rect.h),
                )
            self._parent.window.size = newsize
            diff = self._parent.window.size - newsize
            if self._movewindow == "x" and diff.x != 0:
                posrel.x += diff.x
            elif self._movewindow == "y" and diff.y != 0:
                posrel.y += diff.y
            elif self._movewindow == "xy" and diff.length() != 0:
                posrel += diff
            self._parent.relative = self._parent.window.size - previous
            self._parent.cumulative_relative = (
                self._parent.window.size - self._parent._start_val
            )
            if posrel.length() != 0 and self._parent._resize_winpos is not None:
                self._parent.window.position = self._parent._constrain_position(
                    self._parent._resize_winpos - posrel
                )
            callback = self._parent._callbacks["resize"]["during"]
            if callback is not None:
                callback()


_SNAP_POSITIONS = typing.Literal[
    "top",
    "topleft",
    "left",
    "bottomleft",
    "bottom",
    "bottomright",
    "right",
    "topright",
    "top",
    "height",
]


class CustomWindowBehavior:
    def __init__(
        self,
        window: pygame.Window,
        borders: CustomWindowBorders,
        display_size: typing.Sequence[float],
        taskbar_size: int = 30,
        taskbar_position: typing.Literal["bottom", "left", "right", "top"] = "bottom",
        double_click_cooldown: int | None = 300,
        snap_border_size=10,
        snap_corner_size=30,
        allow_snap: bool = True,
        before_maximize_data: tuple[tuple[int, int], tuple[int, int]] | None = None,
    ):
        self._taskbar_size = taskbar_size
        self._taskbar_position = taskbar_position
        self._display_size = display_size
        self.window = window
        self.borders = borders
        self.allow_snap = allow_snap
        self._maximized = False
        self._before_maximize_data = before_maximize_data
        self._minimized = False
        self._snapped = False
        self._snapped_name = None
        self._snapped_h = False
        self._snap_preview = False
        self._before_snap_data = None
        self._before_snap_preview_size = None
        self._maximize_snap_time = pygame.time.get_ticks()
        self._maximize_time = pygame.time.get_ticks()
        self._fullscreen = False
        self.borders._callbacks["move"]["during"] = self._during_move
        self.borders._callbacks["resize"]["during"] = self._during_resize
        self.borders._callbacks["move"]["end"] = self._end_move
        self.borders._callbacks["resize"]["end"] = self._end_resize
        self.double_click_cooldown = double_click_cooldown
        self.snap_border_size = snap_border_size
        self.snap_corner_size = snap_corner_size
        self._rebuild()

    @property
    def fullscreen(self):
        return self._fullscreen

    @fullscreen.setter
    def fullscreen(self, value):
        if value:
            self.fullscreen_on()
        else:
            self.fullscreen_off()

    @property
    def snapped(self) -> None | str:
        if not self._snapped:
            if self._snapped_name in ["top", "bottom"]:
                return self._snapped_name
            return None
        if self._snapped_h:
            return "height"
        return self._snapped_name

    @snapped.setter
    def snapped(self, value: _SNAP_POSITIONS | None):
        if value is None:
            self.unsnap()
            return
        else:
            self.snap(value)

    @property
    def before_maximized_data(self):
        return self._before_maximize_data

    @property
    def maximized(self):
        return self._maximized

    @maximized.setter
    def maximized(self, v):
        if v:
            self.maximize()
        else:
            self.unmaximize()

    @property
    def minimized(self):
        return self._minimized

    @minimized.setter
    def minimized(self, v):
        if v:
            self.minimize()
        else:
            self.unminimize()

    @property
    def display_size(self):
        return self._display_size

    @display_size.setter
    def display_size(self, value):
        self._display_size = value
        self._rebuild()

    @property
    def taskbar_size(self):
        return self._taskbar_size

    @taskbar_size.setter
    def taskbar_size(self, value):
        self._taskbar_size = value
        self._rebuild()

    @property
    def taskbar_position(self):
        return self._taskbar_position

    @taskbar_position.setter
    def taskbar_position(self, value):
        self._taskbar_position = value
        self._rebuild()

    @property
    def double_click_cooldown(self):
        return (
            self.borders._double_click_cooldown
            if self.borders._toggle_maximize_func is not None
            else None
        )

    @double_click_cooldown.setter
    def double_click_cooldown(self, value):
        if value is None:
            self.borders._toggle_maximize_func = None
        else:
            self.borders._double_click_cooldown = value
            self.borders._toggle_maximize_func = self.toggle_maximize

    def _rebuild(self):
        r = None
        match self._taskbar_position:
            case "left":
                r = pygame.Rect(
                    self._taskbar_size,
                    0,
                    self._display_size[0] - self._taskbar_size,
                    self._display_size[1],
                )
            case "right":
                r = pygame.Rect(
                    0,
                    0,
                    self._display_size[0] - self._taskbar_size,
                    self._display_size[1],
                )
            case "top":
                r = pygame.Rect(
                    0,
                    self._taskbar_size,
                    self._display_size[0],
                    self._display_size[1] - self._taskbar_size,
                )
            case "bottom":
                r = pygame.Rect(
                    0,
                    0,
                    self._display_size[0],
                    self._display_size[1] - self._taskbar_size,
                )
        if r is None:
            r = pygame.Rect()
        self._display_rect = r
        cs = self.snap_corner_size
        bs = self.snap_border_size
        self._snap_areas = {
            "topleft": pygame.Rect(r.left, r.top, cs, cs),
            "left": pygame.Rect(r.left, r.top + cs, bs, r.h - cs * 2),
            "bottomleft": pygame.Rect(r.left, r.top + r.h - cs, cs, cs),
            "bottom": pygame.Rect(r.left + cs, r.top + r.h - cs, r.w - cs * 2, bs),
            "bottomright": pygame.Rect(r.left + r.w - cs, r.top + r.h - cs, cs, cs),
            "right": pygame.Rect(r.left + r.w - bs, r.top + cs, bs, r.h - cs * 2),
            "topright": pygame.Rect(r.left + r.w - cs, r.top, cs, cs),
            "top": pygame.Rect(r.left + cs, r.top, r.w - cs * 2, bs),
        }
        self._snap_rects = {
            "topleft": pygame.Rect(r.left, r.top, r.w / 2, r.h / 2),
            "left": pygame.Rect(r.left, r.top, r.w / 2, r.h),
            "bottomleft": pygame.Rect(r.left, r.top + r.h / 2, r.w / 2, r.h / 2),
            "bottom": pygame.Rect(r.left, r.top, r.w, r.h),
            "bottomright": pygame.Rect(
                r.left + r.w / 2, r.top + r.h / 2, r.w / 2, r.h / 2
            ),
            "right": pygame.Rect(r.left + r.w / 2, r.top, r.w / 2, r.h),
            "topright": pygame.Rect(r.left + r.w / 2, r.top, r.w / 2, r.h / 2),
            "top": pygame.Rect(r.left, r.top, r.w, r.h),
        }

    def center(self):
        w, h = self.window.size
        self.window.position = (
            self._display_rect.w / 2 - w / 2,
            self._display_rect.h / 2 - h / 2,
        )

    def snap(self, position: _SNAP_POSITIONS):
        if not self._snapped:
            self._before_snap_data = self.window.size, self.window.position
        if position == "height":
            self.window.position = (self.window.position[0], self._display_rect.top)
            self.window.size = (self.window.size[0], self._display_rect.h)
            self._snapped_h = True
            self._snapped = True
            return
        self._snapped_name = position
        if position in ["top", "bottom"]:
            self.maximize()
        else:
            rect = self._snap_rects[position]
            self.window.size, self.window.position = rect.size, rect.topleft
            self._snapped = True

    def fullscreen_on(self):
        if self._fullscreen:
            return
        self._maximized = self._snapped = self._snap_preview = self._snapped_h = False
        self._snapped_name = None
        self._before_maximize_data = self.window.size, self.window.position
        self.window.position = (0, 0)
        self.window.size = self._display_size
        self.borders.active = False
        self._fullscreen = True

    def fullscreen_off(self):
        if not self._fullscreen or self._before_maximize_data is None:
            return
        self._fullscreen = False
        self.borders.active = True
        self.window.size, self.window.position = self._before_maximize_data

    def toggle_fullscreen(self):
        if self._fullscreen:
            self.fullscreen_off()
        else:
            self.fullscreen_on()

    def minimize(self):
        if self._minimized:
            return
        self.window.minimize()
        self._minimized = True

    def unminimize(self):
        if not self._minimized:
            return
        self.window.restore()
        self.window.focus()
        self._minimized = True

    def toggle_minimize(self):
        if self._minimized:
            self.unminimize()
        else:
            self.minimize()

    def maximize(self):
        if self._maximized:
            return
        self._maximized = True
        self._before_maximize_data = self.window.size, self.window.position
        self.window.size, self.window.position = (
            self._display_rect.size,
            self._display_rect.topleft,
        )
        self.borders.resizing = self.borders.dragging = False
        self._snapped = self._snapped_h = False
        self._snapped_name = None
        self._maximize_snap_time = self._maximize_time = pygame.time.get_ticks()
        self.borders.active_resize = False

    def unmaximize(self):
        if not self._maximized or self._before_maximize_data is None:
            return
        self._maximized = self._snapped = self._snapped_h = False
        self._snapped_name = None
        self.window.size, self.window.position = self._before_maximize_data
        self.borders.resizing = self.borders.dragging = False
        self.borders.active_resize = True
        self._maximize_snap_time = self._maximize_time = pygame.time.get_ticks()

    def unsnap(self):
        if not self._snapped or self._before_snap_data is None:
            return
        self._snapped = False
        self._snapped_name = None
        if self._snapped_h:
            self.window.size = (self.window.size[0], self._before_snap_data[0][1])
            self.window.position = (
                self.window.position[0],
                self._before_snap_data[1][1],
            )
        else:
            self.window.size, self.window.position = self._before_snap_data
        self._snapped_h = False
        self._maximize_snap_time = pygame.time.get_ticks()

    def toggle_maximize(self):
        self._snapped = self._snapped_h = False
        self._snapped_name = None
        if self._maximized:
            self.unmaximize()
        else:
            self.maximize()

    def _during_resize(self):
        if self.borders.cumulative_relative.length() == 0:
            return
        if self._maximized and self.borders.relative.length() != 0:
            self.unmaximize()
            return
        if not self.allow_snap:
            return
        mouse = pygame.mouse.get_pos(True)
        collided = False
        for name in ["top", "bottom"]:
            area = self._snap_areas[name]
            if area.collidepoint(mouse):
                collided = True
                if not self._snap_preview:
                    self._before_snap_preview_size = self.window.size
                    self._snap_preview = True
                else:
                    self.window.size = (self.window.size[0], self._snap_rects[name].h)
                break
        if (
            not collided
            and self._snap_preview
            and self._before_snap_preview_size is not None
        ):
            self._snap_preview = False
            self.window.size = (self.window.size[0], self._before_snap_preview_size[1])

    def _end_resize(self):
        self._snap_preview = False
        if (
            not self.allow_snap
            or self._before_snap_preview_size is None
            or pygame.time.get_ticks() - self._maximize_time < 100
        ):
            return
        mouse = pygame.mouse.get_pos(True)
        for name in ["top", "bottom"]:
            if self._snap_areas[name].collidepoint(mouse):
                if not self._snapped:
                    self._before_snap_data = (
                        self._before_snap_preview_size,
                        self.window.position,
                    )
                    self._snapped = True
                    self._snapped_h = True
                self.window.size = (self.window.size[0], self._display_rect.h)
                self.window.position = (self.window.position[0], self._display_rect.top)
                self._maximize_snap_time = pygame.time.get_ticks()
                break

    def _during_move(self):
        if self.borders.cumulative_relative.length() == 0:
            return
        if self._maximized and self._before_maximize_data is not None:
            self._maximized = False
            self.window.size, self.window.position = self._before_maximize_data
            pygame.mouse.set_pos(
                (self.window.size[0] / 2, self.borders.titlebar_height / 2)
            )
            self.borders.mouse_changed()
            self._maximize_snap_time = self._maximize_time = pygame.time.get_ticks()
            self.borders.active_resize = True
            return
        if self._snapped:
            self.unsnap()
            pygame.mouse.set_pos(
                (self.window.size[0] / 2, self.borders.titlebar_height / 2)
            )
            self.borders.mouse_changed()
            self._maximize_snap_time = pygame.time.get_ticks()
            return
        if (
            self.allow_snap
            and pygame.time.get_ticks() - self._maximize_snap_time >= 100
        ):
            mouse = pygame.mouse.get_pos(True)
            collided = False
            for name, area in self._snap_areas.items():
                if area.collidepoint(mouse):
                    collided = True
                    if not self._snap_preview:
                        self._before_snap_preview_size = self.window.size
                        self._snap_preview = True
                    else:
                        self.window.size = self._snap_rects[name].size
                    break
            if (
                not collided
                and self._snap_preview
                and self._before_snap_preview_size is not None
            ):
                self._snap_preview = False
                self.window.size = self._before_snap_preview_size

    def _end_move(self):
        if self.borders.cumulative_relative.length() == 0:
            return
        self._snap_preview = False
        if (
            not self.allow_snap
            or self._before_snap_preview_size is None
            or pygame.time.get_ticks() - self._maximize_time < 100
        ):
            return
        mouse = pygame.mouse.get_pos(True)
        for name, area in self._snap_areas.items():
            if area.collidepoint(mouse):
                self._snapped_name = name
                if name in ["top", "bottom"]:
                    if not self._maximized:
                        self._maximized = True
                        self._before_maximize_data = (
                            self._before_snap_preview_size,
                            self.window.position,
                        )
                        self.window.size, self.window.position = (
                            self._display_rect.size,
                            self._display_rect.topleft,
                        )
                        self.borders.resizing = self.borders.dragging = False
                        self._snapped = self._snapped_h = False
                        self._maximize_snap_time = self._maximize_time = (
                            pygame.time.get_ticks()
                        )
                        self.borders.active_resize = False
                else:
                    if not self._snapped:
                        self._before_snap_data = (
                            self._before_snap_preview_size,
                            self.window.position,
                        )
                        self._snapped = True
                        self._snapped_h = False
                        self._maximize_snap_time = pygame.time.get_ticks()
                        rect = self._snap_rects[name]
                        self.window.size = rect.size
                        self.window.position = rect.topleft
                break



import typing
import pygame
import math
from mili import typing as _typing
from mili import error as _error
from mili import data as _data
import mili._core as _core

_animators: list["Animator"] = []

__all__ = (
    "EaseLinear",
    "EaseIn",
    "EaseOut",
    "EaseInCirc",
    "EaseOutCirc",
    "EaseInSin",
    "EaseOutSin",
    "EaseInOutSin",
    "EaseOvershoot",
    "EaseBounce",
    "Animator",
    "ABAnimation",
    "StepsAnimation",
    "StatusAnimation",
    "update_all",
)


class _BaseEase:
    def __call__(self, t: float) -> float:
        return NotImplemented

    def get(self, t: float) -> float:
        return self.__call__(t)


class _BasePowerEase(_BaseEase):
    def __init__(self, n: int = 2):
        if n < 1:
            raise _error.MILIValueError("n argument must be >= 1")
        self.n = n


class EaseLinear(_BaseEase):
    def __call__(self, t):
        return t


class EaseIn(_BasePowerEase):
    def __call__(self, t):
        return t**self.n


class EaseOut(_BasePowerEase):
    def __call__(self, t):
        return 1 - (1 - t) ** self.n


class EaseInCirc(_BasePowerEase):
    def __call__(self, t):
        return 1 - math.sqrt(1 - t**self.n)


class EaseOutCirc(_BasePowerEase):
    def __call__(self, t):
        return math.sqrt(1 - (1 - t) ** self.n)


class EaseInSin(_BasePowerEase):
    def __call__(self, t):
        return 1 - math.cos(t * 0.5 * math.pi) ** (1 / self.n)


class EaseOutSin(_BasePowerEase):
    def __call__(self, t):
        return math.sin(t * 0.5 * math.pi) ** (1 / self.n)


class EaseInOutSin(_BaseEase):
    def __call__(self, t):
        return -math.cos(t * math.pi) / 2 + 0.5


class EaseOvershoot(_BaseEase):
    def __call__(self, t):
        return 1 - math.cos(4 * math.pi * t) * (1 - t)


class EaseBounce(_BaseEase):
    def __call__(self, t):
        if 2 * t - 2 == 0:
            return (1 - t) * abs(math.sin(0))
        return (1 - t) * abs(math.sin(math.pi / (2 * t - 2)))


def update_all():
    for anim in _animators:
        anim.update()


class Animator:
    EPSILON: typing.ClassVar[float] = 0.01

    def __init__(
        self,
        start_value: _typing.AnimValueLike,
        end_value: _typing.AnimValueLike,
        value_type: typing.Literal["number", "color", "sequence"],
        duration_ms: int,
        easing: _typing.EasingLike | None = None,
        finish_callback: typing.Callable[[], float] | None = None,
    ):
        if easing is None:
            easing = EaseLinear()
        self.start_value = start_value
        self.end_value = end_value
        self.value: _typing.AnimValueLike = self.start_value
        self.duration_ms = duration_ms
        self.easing = easing
        self.value_type = value_type
        self.finish_callback = finish_callback
        self.active = False
        self._start_time = 0
        if self not in _animators:
            _animators.append(self)
        if self.duration_ms == 0:
            raise _error.MILIValueError("Duration must be different from 0")
        if self.value_type not in ["number", "color", "sequence"]:
            raise _error.MILIValueError("Unsupported value type")

    def change_values(
        self, start_value: _typing.AnimValueLike, end_value: _typing.AnimValueLike
    ):
        self.start_value, self.end_value = start_value, end_value

    def start(self) -> typing.Self:
        self._start_time = pygame.time.get_ticks()
        self.active = True
        self.value = self.start_value
        return self

    def stop(self) -> typing.Self:
        self.active = False
        return self

    def flip_direction(self) -> typing.Self:
        self.start_value, self.end_value = self.end_value, self.start_value
        return self

    def update(self) -> typing.Self:
        if not self.active:
            return self
        t = self.easing((pygame.time.get_ticks() - self._start_time) / self.duration_ms)
        if self.value_type == "number":
            self.value = pygame.math.lerp(self.start_value, self.end_value, t)  # type: ignore
            if abs(self.value - self.end_value) <= self.EPSILON:  # type: ignore
                self.value = self.end_value
                self.stop()
                if self.finish_callback:
                    self.finish_callback()
        elif self.value_type == "color":
            self.value = self.start_value.lerp(  # type: ignore
                self.end_value,  # type: ignore
                pygame.math.clamp(t, 0, 1),
            )
            if (
                abs(self.value.r - self.end_value.r) <= 1  # type: ignore
                and abs(self.value.g - self.end_value.g) <= 1  # type: ignore
                and abs(self.value.b - self.end_value.b) <= 1  # type: ignore
                and abs(self.value.a - self.end_value.a) <= 1  # type: ignore
            ):
                self.value = self.end_value
                self.stop()
                if self.finish_callback:
                    self.finish_callback()
        elif self.value_type == "sequence":
            cur_sequence = []
            finished = 0
            for sval, eval in zip(self.start_value, self.end_value):  # type: ignore
                cval = pygame.math.lerp(sval, eval, t)
                if abs(cval - eval) <= self.EPSILON:
                    cval = eval
                    finished += 1
                cur_sequence.append(cval)
            self.value = cur_sequence
            if finished >= len(self.start_value):  # type: ignore
                self.value = self.end_value
                self.stop()
                if self.finish_callback:
                    self.finish_callback()
        else:
            raise _error.MILIValueError(f"Unsupported value type '{self.value_type}'")
        return self


class _AnimatorWrapper:
    animator: Animator

    def stop(self) -> typing.Self:
        self.animator.stop()
        return self

    def update(self) -> typing.Self:
        self.animator.update()
        return self

    @property
    def value(self) -> _typing.AnimValueLike:
        return self.animator.value

    @value.setter
    def value(self, value):
        self.animator.value = value

    @property
    def active(self) -> bool:
        return self.animator.active


class ABAnimation(_AnimatorWrapper):
    def __init__(
        self,
        a: _typing.AnimValueLike,
        b: _typing.AnimValueLike,
        value_type: typing.Literal["number", "color", "sequence"],
        ab_duration_ms: int,
        ba_duration_ms: int,
        easing: _typing.EasingLike | None = None,
        reach_a_callback: typing.Callable[[], typing.Any] | None = None,
        reach_b_callback: typing.Callable[[], typing.Any] | None = None,
    ):
        self.a, self.b = a, b
        self.direction: str = "ab"
        self.ab_duration_ms, self.ba_duration_ms = ab_duration_ms, ba_duration_ms
        self.reach_a_callback, self.reach_b_callback = (
            reach_a_callback,
            reach_b_callback,
        )
        self.animator = Animator(
            self.a,
            self.b,
            value_type,
            self.ab_duration_ms,
            easing,
            self.reach_b_callback,
        )

    def goto_a(self) -> typing.Self:
        self.animator.change_values(self.animator.value, self.a)
        self.animator.finish_callback = self.reach_a_callback
        self.direction = "ba"
        self.animator.start()
        return self

    def goto_b(self) -> typing.Self:
        self.animator.change_values(self.animator.value, self.b)
        self.animator.finish_callback = self.reach_b_callback
        self.direction = "ab"
        self.animator.start()
        return self

    def flip_direction(self) -> typing.Self:
        if self.direction == "ba":
            self.goto_b()
        else:
            self.goto_a()
        return self

    def change_ab(
        self, a: _typing.AnimValueLike, b: _typing.AnimValueLike
    ) -> typing.Self:
        self.a, self.b = a, b
        if self.direction == "ab":
            self.animator.end_value = self.b
        else:
            self.animator.end_value = self.a
        return self


class StepsAnimation(_AnimatorWrapper):
    def __init__(
        self,
        steps: typing.Sequence[_typing.AnimValueLike],
        value_type: typing.Literal["number", "color", "sequence"],
        duration_ms: int,
        easing: _typing.EasingLike | None = None,
        reach_callback: typing.Callable[[], typing.Any] | None = None,
    ):
        if len(steps) <= 2:
            raise _error.MILIValueError("At least 2 animation steps are required")
        self.steps = steps
        self.step = -1
        self.animator = Animator(
            self.steps[0],
            self.steps[1],
            value_type,
            duration_ms,
            easing,
            reach_callback,
        )

    def goto(self, step: int) -> typing.Self:
        if step < 0 or step >= len(self.steps):
            raise _error.MILIValueError("Step out of bounds")
        self.step = step
        self.animator.change_values(self.animator.value, self.steps[step])
        self.animator.start()
        return self


class StatusAnimation(_AnimatorWrapper):
    def __init__(
        self,
        base_value: _typing.AnimValueLike,
        hover_value: _typing.AnimValueLike | None,
        press_value: _typing.AnimValueLike | None,
        value_type: typing.Literal["number", "color"],
        duration_ms: int,
        easing: _typing.EasingLike | None = None,
        reach_callback: typing.Callable[[], typing.Any] | None = None,
        style_name: str = "",
        press_button: int = pygame.BUTTON_LEFT,
        update_id: str | None = None,
    ):
        self.base_value = base_value
        self.hover_value = hover_value
        self.press_value = press_value
        self.press_button = press_button
        self.style_name = style_name
        self._last_value = self.base_value
        self.animator = Animator(
            base_value,
            base_value,
            value_type,
            duration_ms,
            easing,
            reach_callback,
        )
        _core._globalctx._register_update_id(update_id, self.status_update)

    def status_update(
        self, element: _data.Interaction, selected: bool = False
    ) -> _data.Interaction:
        if element.hovered:
            if self.press_value is not None and (
                element.press_button == self.press_button or selected
            ):
                if self._last_value != self.press_value:
                    self.animator.start_value = self._last_value
                    self._last_value = self.press_value
                    self.animator.end_value = self.press_value
                    self.animator.start()
            elif self.hover_value is not None:
                if self._last_value != self.hover_value:
                    self.animator.start_value = self._last_value
                    self._last_value = self.hover_value
                    self.animator.end_value = self.hover_value
                    self.animator.start()
            elif self._last_value != self.base_value:
                self.animator.start_value = self._last_value
                self._last_value = self.base_value
                self.animator.end_value = self.base_value
                self.animator.start()
        else:
            if (
                selected or element.press_button == self.press_button
            ) and self.press_value is not None:
                if self._last_value != self.press_value:
                    self.animator.start_value = self._last_value
                    self._last_value = self.press_value
                    self.animator.end_value = self.press_value
                    self.animator.start()
            elif self._last_value != self.base_value:
                self.animator.start_value = self._last_value
                self._last_value = self.base_value
                self.animator.end_value = self.base_value
                self.animator.start()
        return element

    @property
    def style(self):
        return {self.style_name: self.value}



import pygame
import typing
from dataclasses import dataclass

if typing.TYPE_CHECKING:
    from mili import MILI as _MILI

__all__ = (
    "ImageCache",
    "TextCache",
    "ParentCache",
    "ImageLayerCache",
    "Interaction",
    "ElementData",
)


class ParentCache:
    def __init__(self, static):
        self._cache = None
        self._rebuild = True
        self._size = None
        self._grid = None
        self._static = static
        self._style = None

    def refresh(self):
        self._rebuild = True
        self._cache = {
            "children": [],
            "children_grid": [],
            "children_fillx": [],
            "children_filly": [],
        }
        self._grid = None


class ImageCache:
    _preallocated_caches = []
    _preallocated_index = -1

    def __init__(self):
        self._cache: dict[str, typing.Any] | None = None

    def get_output(self) -> pygame.Surface | None:
        if self._cache is None:
            return None
        return self._cache.get("output", None)

    @classmethod
    def get_next_cache(cls):
        cls._preallocated_index += 1
        if cls._preallocated_index > len(cls._preallocated_caches) - 1:
            cls._preallocated_caches.append(ImageCache())
        return cls._preallocated_caches[cls._preallocated_index]


class ImageLayerCache:
    def __init__(
        self,
        mili: "_MILI",
        size: typing.Sequence[float],
        offset: typing.Sequence[float] = (0, 0),
        blit_flags: int = 0,
    ):
        self._offset = pygame.Vector2(0, 0)
        self._surface = pygame.Surface((0, 0))
        self.size = size
        self.active = True
        self.blit_flags: int = blit_flags
        self._offset = pygame.Vector2(offset)
        self._erase_rects: tuple[pygame.Rect, ...] | None = None
        self._mili = mili
        self._caches = []
        self._caches_set = set()
        self._caches_activity = {}
        self._rendered = False
        self._mili._ctx._image_layer_caches.append(self)

    @property
    def erase_rects(self) -> tuple[pygame.Rect, ...] | None:
        return self._erase_rects

    @erase_rects.setter
    def erase_rects(self, v):
        if v is None:
            if v != self._erase_rects:
                self._dirty = True
            self._erase_rects = v
            return
        v = tuple(v)
        if self._erase_rects is None:
            self._erase_rects = v
            self._dirty = True
            return
        if len(v) != len(self._erase_rects):
            self._erase_rects = v
            self._dirty = True
            return
        for i in range(len(v) - 1):
            a, b = v[i], self._erase_rects[i]
            if a != b:
                self._dirty = True
                break
        self._erase_rects = v

    @property
    def offset(self) -> pygame.Vector2:
        return self._offset

    @offset.setter
    def offset(self, v: typing.Sequence[float]):
        if v == self._offset:
            return
        self._offset = pygame.Vector2(v)
        self._dirty = True

    @property
    def size(self) -> pygame.Vector2:
        return pygame.Vector2(self._surface.size)

    @size.setter
    def size(self, v: typing.Sequence[float]):
        if v == self._surface.size:
            return
        self._surface = pygame.Surface(v, pygame.SRCALPHA)
        self._dirty = True

    def destroy(self):
        self._mili._ctx._image_layer_caches.remove(self)


class TextCache:
    _preallocated_caches = []
    _preallocated_index = -1

    def __init__(self):
        self._cache: dict[str, typing.Any] | None = None
        self._rich: dict | None = None

    def get_output(self) -> pygame.Surface | None:
        if self._cache is None:
            return None
        return self._cache["output"]

    @classmethod
    def get_next_cache(cls):
        cls._preallocated_index += 1
        if cls._preallocated_index > len(cls._preallocated_caches) - 1:
            cls._preallocated_caches.append(TextCache())
        return cls._preallocated_caches[cls._preallocated_index]


@dataclass(slots=True)
class ElementGridData:
    overflowx: float
    overflowy: float
    padx: float
    pady: float
    spacing: float
    spacex: float
    spacey: float


@dataclass
class ElementData:
    rect: pygame.Rect
    absolute_rect: pygame.Rect
    z: int
    id: int
    style: dict[str, typing.Any]
    components: dict[typing.Literal["type", "style", "data"], typing.Any]
    parent_id: int
    _grid: dict | None | ElementGridData
    _children: list[dict]
    _children_ids: list[int] | None = None

    @property
    def grid(self) -> ElementGridData:
        if isinstance(self._grid, ElementGridData):
            return self._grid
        if self._grid is None:
            self._grid = ElementGridData(0, 0, 0, 0, 0, 0, 0)
            return self._grid
        self._grid = ElementGridData(**self._grid)
        return self._grid

    @property
    def children_ids(self) -> list[int]:
        if self._children_ids is None:
            self._children_ids = [el["id"] for el in self._children]
        return self._children_ids


@dataclass
class Interaction:
    _mili: "_MILI"
    hovered: bool
    press_button: int
    just_pressed_button: int
    just_released_button: int
    absolute_hover: bool
    unhover_pressed: bool
    _just_hover: int
    _raw_data: dict[str, typing.Any]
    _data: ElementData | None = None
    _did_begin: bool = False
    parent: "Interaction|None" = None

    @property
    def just_hovered(self) -> bool:
        return self._just_hover == 1

    @property
    def just_unhovered(self) -> bool:
        return self._just_hover == -1

    @property
    def left_pressed(self) -> bool:
        return self.press_button == pygame.BUTTON_LEFT

    @property
    def left_just_pressed(self) -> bool:
        return self.just_pressed_button == pygame.BUTTON_LEFT

    @property
    def left_just_released(self) -> bool:
        return self.just_released_button == pygame.BUTTON_LEFT

    left_clicked = left_just_released

    @property
    def right_pressed(self) -> bool:
        return self.press_button == pygame.BUTTON_RIGHT

    @property
    def right_just_pressed(self) -> bool:
        return self.just_pressed_button == pygame.BUTTON_RIGHT

    @property
    def right_just_released(self) -> bool:
        return self.just_released_button == pygame.BUTTON_RIGHT

    right_clicked = right_just_released

    @property
    def middle_pressed(self) -> bool:
        return self.press_button == pygame.BUTTON_MIDDLE

    @property
    def middle_just_pressed(self) -> bool:
        return self.just_pressed_button == pygame.BUTTON_MIDDLE

    @property
    def middle_just_released(self) -> bool:
        return self.just_released_button == pygame.BUTTON_MIDDLE

    middle_clicked = middle_just_released

    @property
    def data(self) -> ElementData:
        if self._data is None:
            self._data = self._mili._ctx._coreutils._element_data(
                self._mili._ctx, self._raw_data
            )
        return self._data

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        if self._did_begin:
            self._mili.end()



import pygame
import typing
import math
from mili import error
from mili import data as _data

if typing.TYPE_CHECKING:
    from mili._core import _ctx

__all__ = ()

_VALID_ALIGN = set(["first", "last", "center"])
_VALID_ANCHOR = set(["first", "last", "center", "max_spacing"])
_VALID_ALIGN_GRID = set(["first", "last", "center", "first_center", "last_center"])
_OUTLINE_OFFSETS = [(0, 0), (2, 0), (0, 2), (2, 2), (1, 0), (1, 2), (0, 1), (2, 1)]
_OUTLINE_OFFSETS_CENTERED = [
    (offset[0] - 1, offset[1] - 1) for offset in _OUTLINE_OFFSETS
]
_FONT_ALIGNS = {
    "left": pygame.FONT_LEFT,
    "center": pygame.FONT_CENTER,
    "right": pygame.FONT_RIGHT,
    pygame.FONT_LEFT: pygame.FONT_LEFT,
    pygame.FONT_CENTER: pygame.FONT_CENTER,
    pygame.FONT_RIGHT: pygame.FONT_RIGHT,
}
_FONT_DIRS = {
    "ltr": pygame.DIRECTION_LTR,
    "rtl": pygame.DIRECTION_RTL,
    "ttb": pygame.DIRECTION_TTB,
    "btt": pygame.DIRECTION_BTT,
    pygame.DIRECTION_LTR: pygame.DIRECTION_LTR,
    pygame.DIRECTION_RTL: pygame.DIRECTION_RTL,
    pygame.DIRECTION_TTB: pygame.DIRECTION_TTB,
    pygame.DIRECTION_BTT: pygame.DIRECTION_BTT,
}
_number_mods = {}


def _render_layer_cache(self: _data.ImageLayerCache, canva: pygame.Surface | None):
    if canva is None or not self.active or self._rendered:
        return
    remove = []
    if self._dirty:
        self._surface.fill(0)
        for cache in self._caches:
            if not self._caches_activity.get(cache, False):
                remove.append(cache)
                continue
            self._surface.blit(
                cache._cache["output"], cache._cache["pos"] - self._offset
            )
        if self._erase_rects is not None:
            for rect in self._erase_rects:
                r = rect.copy()
                r.topleft -= self._offset
                self._surface.fill((0, 0, 0, 0), r)
        self._dirty = False
    else:
        for cache in self._caches:
            if not self._caches_activity.get(cache, False):
                remove.append(cache)
    for cache in remove:
        self._caches.remove(cache)
        self._caches_set.remove(cache)
    canva.blit(self._surface, self._offset, special_flags=self.blit_flags)
    self._rendered = True


def _abs_perc(val, dim) -> float:
    if isinstance(val, str):
        val = val.strip()
        sign = 1
        add = 0
        addsign = 1
        if val.startswith("-"):
            val = val.removeprefix("-")
            sign = -1
        for c in ["-", "+"]:
            if c in val:
                val, add = val.split(c)
                val = val.strip()
                add = add.strip()
                try:
                    add = float(add)
                except ValueError:
                    raise error.MILIValueError(
                        f"Invalid percentage addition: '{c}{add}'"
                    )
                addsign = 1 if c == "+" else -1
        if val.endswith("%"):
            val = val.removesuffix("%")
        elif not (key := val[0]).isnumeric():
            if key in _number_mods:
                number = float(val.removeprefix(key))
                return _number_mods[key](number) * sign + add * addsign
            else:
                raise error.MILIValueError(
                    f"The character prefix '{key}' of the value '{val}' is not registered as a number modifier and cannot be evaluated"
                )
        try:
            val = float(val)
        except ValueError:
            raise error.MILIValueError(f"Invalid percentage value '{val}'")
        return ((dim * float(val)) / 100) * sign + add * addsign
    return val


def _abs_mod(val) -> float:
    if isinstance(val, str):
        val = val.strip()
        sign = 1
        if val.startswith("-"):
            val = val.removeprefix("-")
            sign = -1
        if not (key := val[0]).isnumeric():
            if key in _number_mods:
                number = float(val.removeprefix(key))
                return _number_mods[key](number) * sign
            else:
                raise error.MILIValueError(
                    f"The character prefix '{key}' of the value '{val}' is not registered as a number modifier and cannot be evaluated"
                )
        try:
            val = float(val)
        except ValueError:
            raise error.MILIValueError(f"Invalid number in string '{val}'")
        return val * sign
    return val


def _get_first_button(buttons):
    bid = -1
    for i, b in enumerate(buttons):
        if b:
            bid = i + 1
            break
    return bid


def _nine_patch(image: pygame.Surface, w, h, s, smoothscale):
    ow, oh = image.get_size()
    scale_func = pygame.transform.smoothscale if smoothscale else pygame.transform.scale
    subsurf = image.subsurface
    ohsize, hsize = ow - s * 2, w - s * 2
    ovsize, vsize = oh - s * 2, h - s * 2
    base = pygame.Surface((w, h), pygame.SRCALPHA)
    base.blit(subsurf((0, 0, s, s)), (0, 0))  # topleft
    base.blit(subsurf((ow - s, 0, s, s)), (w - s, 0))  # topright
    base.blit(subsurf((0, oh - s, s, s)), (0, h - s))  # bottomleft
    base.blit(subsurf((ow - s, oh - s, s, s)), (w - s, h - s))  # bottomright
    if ohsize > 0:
        base.blit(scale_func(subsurf((s, 0, ohsize, s)), (hsize, s)), (s, 0))  # top
        base.blit(
            scale_func(subsurf((s, oh - s, ohsize, s)), (hsize, s)),
            (s, h - s),
        )  # bottom
    if ovsize > 0:
        base.blit(scale_func(subsurf((0, s, s, ovsize)), (s, vsize)), (0, s))  # left
        base.blit(
            scale_func(subsurf((ow - s, s, s, ovsize)), (s, vsize)),
            (w - s, s),
        )  # right
    if ohsize > 0 and ovsize > 0:
        base.blit(
            scale_func(subsurf((s, s, ohsize, ovsize)), (hsize, vsize)),
            (s, s),
        )  # inner
    return base


def _render_text(
    data,
    font: pygame.Font,
    fontalign,
    fontdir,
    antialias,
    color,
    bg_color,
    bold,
    italic,
    underline,
    strikethrough,
    wraplen,
    outline_color,
):
    font.align = fontalign
    font.bold = bold
    font.italic = italic
    font.underline = underline
    font.strikethrough = strikethrough
    font.set_direction(fontdir)
    if outline_color is None:
        return font.render(str(data), antialias, color, bg_color, max(0, int(wraplen)))
    middle = font.render(str(data), antialias, color, None, max(0, int(wraplen)))
    outline = font.render(
        str(data), antialias, outline_color, bg_color, max(0, int(wraplen))
    )
    output = pygame.Surface((middle.width + 2, middle.height + 2), pygame.SRCALPHA)
    for offset in _OUTLINE_OFFSETS:
        output.blit(outline, offset)
    output.blit(middle, (1, 1))
    return output


def _get_image(
    rect: pygame.Rect,
    data: pygame.Surface,
    padx,
    pady,
    do_fill,
    do_stretchx,
    do_stretchy,
    fill_color,
    border_radius,
    alpha,
    smoothscale,
    nine_patch,
    transforms,
    filters,
) -> pygame.Surface:
    if transforms is not None:
        for transform in transforms:
            if callable(transform):
                res = transform(data)
            else:
                func, *args = transform
                res = func(data, *args)
            if not isinstance(res, pygame.Surface):
                raise error.MILIValueError(
                    "Image transform must return a new Surface or the same Surface"
                )
            data = res
    iw, ih = data.size
    tw, th = max(rect.w - padx * 2, 1), max(rect.h - pady * 2, 1)
    if nine_patch == 0:
        if not do_fill:
            w = tw
            h = int(ih * (w / iw))
            if h > th:
                h = th
                w = int(iw * (h / ih))
            w, h = max(w, 1), max(h, 1)
            if w < tw and do_stretchx:
                w = tw
            if h < th and do_stretchy:
                h = th
            if smoothscale:
                image = pygame.transform.smoothscale(data, (w, h))
            else:
                image = pygame.transform.scale(data, (w, h))
        else:
            w = tw
            h = int(ih * (w / iw))
            if h < th:
                h = th
                w = int(iw * (h / ih))
            if smoothscale:
                image = pygame.transform.smoothscale(data, (w, h))
            else:
                image = pygame.transform.scale(data, (w, h))
            image = image.subsurface((w // 2 - tw // 2, h // 2 - th // 2, tw, th))
    else:
        image = _nine_patch(data, tw, th, nine_patch, smoothscale)

    w, h = image.size
    if fill_color is not None:
        image.fill(fill_color)
    if not isinstance(border_radius, float | int) or border_radius > 0:
        mask_surf = pygame.Surface((w, h), pygame.SRCALPHA)
        if isinstance(border_radius, float | int):
            border_radius = int(border_radius)
            if w == h and border_radius == w // 2:
                if smoothscale:
                    pygame.draw.aacircle(
                        mask_surf, (255, 255, 255, 255), (w // 2, h // 2), w // 2
                    )
                else:
                    pygame.draw.circle(
                        mask_surf, (255, 255, 255, 255), (w // 2, h // 2), w // 2
                    )
            else:
                pygame.draw.rect(
                    mask_surf, (255, 255, 255, 255), (0, 0, w, h), 0, border_radius
                )
        else:
            pygame.draw.rect(
                mask_surf, (255, 255, 255, 255), (0, 0, w, h), 0, 0, *border_radius
            )
        image.blit(mask_surf, (0, 0), special_flags=pygame.BLEND_RGBA_MULT)
    if alpha != 255:
        image.set_alpha(alpha)
    if filters is not None:
        for ifilter in filters:
            if callable(ifilter):
                res = ifilter(image)
            else:
                func, *args = ifilter
                res = func(image, *args)
            if not isinstance(res, pygame.Surface):
                raise error.MILIValueError(
                    "Image filter must return a new Surface or the same Surface"
                )
            image = res
    return image


def _element_data(ctx: "_ctx", el: dict[str, typing.Any]) -> "_data.ElementData":
    if el["id"] in ctx._old_data:
        old_data = ctx._old_data[el["id"]]
    else:
        old_data = {
            "rect": pygame.Rect(),
            "abs_rect": pygame.Rect(),
            "children": [],
            "components": [],
            "parent_id": 0,
            "grid": None,
        }
    grid = old_data["grid"]
    return _data.ElementData(
        rect=old_data["rect"].copy(),
        absolute_rect=old_data["abs_rect"].copy(),
        z=el["z"],
        id=el["id"],
        style=el["style"],
        _children=old_data["children"],
        parent_id=old_data["parent_id"],
        components=old_data["components"],
        _grid=grid,
    )


def _align_rect(align, rect: pygame.Rect, padx, pady):
    x, y = 0, 0
    if align == "center":
        x, y = rect.centerx, rect.centery
    elif align == "topleft":
        x, y = rect.x + padx, rect.y + pady
    elif align == "bottomleft":
        x, y = rect.x + padx, rect.bottom - pady
    elif align == "bottomright":
        x, y = rect.right - padx, rect.bottom - pady
    elif align == "topright":
        x, y = rect.right - padx, rect.y + pady
    elif align == "midleft" or align == "left":
        align = "midleft"
        x, y = rect.x + padx, rect.centery
    elif align == "midright" or align == "right":
        align = "midright"
        x, y = rect.right - padx, rect.centery
    elif align == "midtop" or align == "top":
        align = "midtop"
        x, y = rect.centerx, rect.y + pady
    elif align == "midbottom" or align == "bottom":
        align = "midbottom"
        x, y = rect.centerx, rect.bottom - pady
    else:
        raise error.MILIValueError("Invalid text alignment")
    return {align: (x, y)}


def _check_align(align):
    if align not in _VALID_ALIGN:
        didyoumean = ""
        if align == "left" or align == "top":
            didyoumean = ". Did you mean 'first'?"
        elif align == "right" or align == "bottom":
            didyoumean = ". Did you mean 'last'?"
        raise error.MILIValueError(f"Invalid alignment '{align}'{didyoumean}")


def _check_anchor(align):
    if align not in _VALID_ANCHOR:
        didyoumean = ""
        if align == "left" or align == "top":
            didyoumean = ". Did you mean 'first'?"
        elif align == "right" or align == "bottom":
            didyoumean = ". Did you mean 'last'?"
        raise error.MILIValueError(f"Invalid anchoring '{align}'{didyoumean}")


def _check_align_grid(align):
    if align not in _VALID_ALIGN_GRID:
        didyoumean = ""
        if align == "left" or align == "top":
            didyoumean = ". Did you mean 'first'?"
        elif align == "right" or align == "bottom":
            didyoumean = ". Did you mean 'last'?"
        raise error.MILIValueError(f"Invalid grid alignment '{align}'{didyoumean}")


def _partial_interaction(ctx, element, absolute_hover):
    element["hovered"] = False
    if (eid := element["id"]) in ctx._interaction_cache:
        i = ctx._interaction_cache[eid]
        (
            i.hovered,
            i.press_button,
            i.just_pressed_button,
            i.just_released_button,
            i.absolute_hover,
            i.unhover_pressed,
            i._just_hover,
            i._raw_data,
            i._data,
        ) = (False, -1, -1, -1, absolute_hover, False, 0, element, None)
        return i

    i = _data.Interaction(
        ctx._mili,
        False,
        -1,
        -1,
        -1,
        absolute_hover,
        False,
        0,
        element,
    )
    ctx._interaction_cache[eid] = i
    return i


def _total_interaction(ctx, element, absolute_hover, hovered, unhovered, old_el):
    if (eid := element["id"]) in ctx._interaction_cache:
        i = ctx._interaction_cache[eid]
        (
            i.hovered,
            i.press_button,
            i.just_pressed_button,
            i.just_released_button,
            i.absolute_hover,
            i.unhover_pressed,
            i._just_hover,
            i._raw_data,
            i._data,
        ) = (
            absolute_hover and hovered,
            _get_first_button(pygame.mouse.get_pressed(5, ctx._global_mouse)),
            _get_first_button(ctx._get_just_pressed_func()),
            _get_first_button(ctx._get_just_released_func()) if absolute_hover else -1,
            absolute_hover,
            unhovered,
            0,
            old_el,
            None,
        )
    else:
        i = _data.Interaction(
            ctx._mili,
            absolute_hover and hovered,
            _get_first_button(pygame.mouse.get_pressed(5, ctx._global_mouse)),
            _get_first_button(ctx._get_just_pressed_func()),
            _get_first_button(ctx._get_just_released_func()) if absolute_hover else -1,
            absolute_hover,
            unhovered,
            0,
            old_el,
        )
        ctx._interaction_cache[eid] = i
    if i.hovered and not old_el["hovered"]:
        i._just_hover = 1
    if not i.hovered and old_el["hovered"]:
        i._just_hover = -1
    element["hovered"] = i.hovered
    return i


def _border_radius(style, w, h):
    if isinstance(style, (int, float, str)):
        return int(_abs_perc(style, min(w, h)))
    else:
        return [int(_abs_perc(br, min(w, h))) for br in style]


def _get_conditional_color(it, color):
    if it.left_pressed:
        return color["press"]
    if it.hovered:
        return color["hover"]
    return color["default"]


def _draw_line(canva, antialias, width, color, points, dash_size, dash_offset):
    if dash_size is None:
        if antialias and (width == 1 or pygame.vernum >= (2, 5, 6)):
            if pygame.vernum >= (2, 5, 6):
                pygame.draw.aaline(
                    canva,
                    color,
                    points[0],
                    points[1],
                    width,
                )
            else:
                pygame.draw.aaline(
                    canva,
                    color,
                    points[0],
                    points[1],
                )
        else:
            pygame.draw.line(canva, color, points[0], points[1], width)
    else:
        if isinstance(dash_size, (str, int, float)):
            fill_size = space_size = dash_size
        else:
            fill_size, space_size = dash_size

        p1 = pygame.Vector2(points[0])
        p2 = pygame.Vector2(points[1])
        line_len = p1.distance_to(p2)
        fill_size = _abs_perc(fill_size, line_len)
        space_size = _abs_perc(space_size, line_len)
        dash_offset = _abs_perc(dash_offset, line_len)
        if dash_offset >= line_len:
            return
        direction = p2 - p1
        if direction.magnitude() != 0:
            direction.normalize_ip()

        cur_point = p1 + direction * dash_offset
        do_fill = True
        cur_size = fill_size
        while True:
            if cur_size == 0:
                break
            do_break = False
            next_point = cur_point + cur_size * direction
            next_len = p1.distance_to(next_point)
            if next_len >= line_len:
                next_point = p2.copy()
                do_break = True
            if do_fill:
                if antialias and (width == 1 or pygame.vernum >= (2, 5, 6)):
                    if pygame.vernum >= (2, 5, 6):
                        pygame.draw.aaline(canva, color, cur_point, next_point, width)
                    else:
                        pygame.draw.aaline(canva, color, cur_point, next_point)
                else:
                    pygame.draw.line(canva, color, cur_point, next_point, width)
                do_fill = False
                cur_size = space_size
            else:
                do_fill = True
                cur_size = fill_size
            cur_point = next_point
            if do_break:
                break


def _draw_circle(
    canva, circle_rect, antialias, color, outline, corners, dash_size, dash_anchor
):
    if dash_size is None or outline <= 0:
        if circle_rect.w == circle_rect.h:
            if corners is None:
                (pygame.draw.aacircle if antialias else pygame.draw.circle)(
                    canva,
                    color,
                    circle_rect.center,
                    circle_rect.w / 2,
                    outline,
                )
            else:
                (pygame.draw.aacircle if antialias else pygame.draw.circle)(
                    canva,
                    color,
                    circle_rect.center,
                    circle_rect.w / 2,
                    outline,
                    *corners,
                )
        else:
            pygame.draw.ellipse(canva, color, circle_rect, outline)
    else:
        pi2 = math.pi * 2
        if isinstance(dash_size, (int, float)):
            fill_size = space_size = dash_size
        else:
            fill_size, space_size = dash_size
        if isinstance(dash_anchor, (int, float)):
            start_ang = end_ang = dash_anchor
        else:
            start_ang, end_ang = dash_anchor

        start_ang = (start_ang / 100) * pi2
        end_ang = (end_ang / 100) * pi2
        fill_size = (fill_size / 100) * pi2
        space_size = (space_size / 100) * pi2
        real_end = end_ang
        if start_ang >= end_ang:
            real_end = pi2 + end_ang

        cur_ang = start_ang
        do_fill = True
        cur_size = fill_size
        while True:
            if cur_size == 0:
                break
            do_break = False
            next_ang = cur_ang + cur_size
            if next_ang >= real_end:
                next_ang = real_end
                do_break = True
            if do_fill:
                pygame.draw.arc(canva, color, circle_rect, cur_ang, next_ang, outline)
                do_fill = False
                cur_size = space_size
            else:
                do_fill = True
                cur_size = fill_size
            cur_ang = next_ang
            if do_break:
                break


def _dash_get_side(pos, rect):
    if pos <= rect.w:
        return "t", pos
    elif pos <= rect.w + rect.h:
        return "r", pos - rect.w
    elif pos <= rect.w * 2 + rect.h:
        return "b", pos - rect.w - rect.h
    elif pos <= rect.w * 2 + rect.h * 2:
        return "l", pos - rect.w * 2 - rect.h
    else:
        return "o", pos - (rect.w * 2 + rect.h * 2)
    return "", 0


def _draw_rect(canva, color, base_rect, outline, border_radius, dash_size, dash_offset):
    if dash_size is None or outline <= 0:
        if isinstance(border_radius, list):
            pygame.draw.rect(
                canva,
                color,
                base_rect,
                int(outline),
                0,
                *border_radius,
            )
        else:
            pygame.draw.rect(
                canva,
                color,
                base_rect,
                int(outline),
                int(border_radius),
            )
    else:
        if isinstance(dash_size, (str, int, float)):
            fill_size = space_size = dash_size
        else:
            fill_size, space_size = dash_size

        perimeter = base_rect.w * 2 + base_rect.h * 2
        fill_size = _abs_perc(fill_size, perimeter)
        space_size = _abs_perc(space_size, perimeter)
        dash_offset = _abs_perc(dash_offset, perimeter)

        cur_len = dash_offset
        do_draw = True
        step_len = fill_size
        i = 0

        while i < perimeter + 1:
            do_break = False
            next_len = cur_len + step_len
            if next_len > perimeter + dash_offset:
                next_len = perimeter + dash_offset
                do_break = True
            if do_draw:
                a_side, ar = _dash_get_side(cur_len, base_rect)
                b_side, br = _dash_get_side(next_len, base_rect)
                func = _dash_FUNCS[f"{a_side}{b_side}"]
                func(canva, ar, br, base_rect, color, outline)
                step_len = space_size
                do_draw = False
            else:
                step_len = fill_size
                do_draw = True
            cur_len = next_len
            i += 1
            if do_break:
                break


def _dash_tt(canva, a, b, rect, color, size):
    pygame.draw.rect(canva, color, (rect.x + a, rect.y, b - a, size))


def _dash_tr(canva, a, b, rect, color, size):
    _dash_tt(canva, a, rect.w, rect, color, size)
    _dash_rr(canva, 0, b, rect, color, size)


def _dash_tb(canva, a, b, rect, color, size):
    _dash_tt(canva, a, rect.w, rect, color, size)
    _dash_rr(canva, 0, rect.h, rect, color, size)
    _dash_bb(canva, 0, b, rect, color, size)


def _dash_tl(canva, a, b, rect, color, size):
    _dash_tt(canva, a, rect.w, rect, color, size)
    _dash_rr(canva, 0, rect.h, rect, color, size)
    _dash_bb(canva, 0, rect.w, rect, color, size)
    _dash_ll(canva, 0, b, rect, color, size)


def _dash_to(canva, a, b, rect, color, size):
    _dash_tt(canva, a, rect.w, rect, color, size)
    _dash_rr(canva, 0, rect.h, rect, color, size)
    _dash_bb(canva, 0, rect.w, rect, color, size)
    _dash_ll(canva, 0, rect.h, rect, color, size)
    _dash_tt(canva, 0, b, rect, color, size)


def _dash_rr(canva, a, b, rect, color, size):
    pygame.draw.rect(canva, color, (rect.right - size, rect.y + a, size, b - a))


def _dash_rb(canva, a, b, rect, color, size):
    _dash_rr(canva, a, rect.h, rect, color, size)
    _dash_bb(canva, 0, b, rect, color, size)


def _dash_rl(canva, a, b, rect, color, size):
    _dash_rr(canva, a, rect.h, rect, color, size)
    _dash_bb(canva, 0, rect.w, rect, color, size)
    _dash_ll(canva, 0, b, rect, color, size)


def _dash_ro(canva, a, b, rect, color, size):
    _dash_rr(canva, a, rect.h, rect, color, size)
    _dash_bb(canva, 0, rect.w, rect, color, size)
    _dash_ll(canva, 0, rect.h, rect, color, size)
    _dash_tt(canva, 0, b, rect, color, size)


def _dash_bb(canva, a, b, rect, color, size):
    pygame.draw.rect(
        canva, color, (rect.x + (rect.w - b), rect.bottom - size, b - a, size)
    )


def _dash_bl(canva, a, b, rect, color, size):
    _dash_bb(canva, a, rect.w, rect, color, size)
    _dash_ll(canva, 0, b, rect, color, size)


def _dash_bo(canva, a, b, rect, color, size):
    _dash_bb(canva, a, rect.w, rect, color, size)
    _dash_ll(canva, 0, rect.h, rect, color, size)
    _dash_tt(canva, 0, b, rect, color, size)


def _dash_ll(canva, a, b, rect, color, size):
    pygame.draw.rect(canva, color, (rect.x, rect.y + (rect.h - b), size, b - a))


def _dash_lo(canva, a, b, rect, color, size):
    _dash_ll(canva, a, rect.h, rect, color, size)
    _dash_tt(canva, 0, b, rect, color, size)


_dash_FUNCS = {
    "tt": _dash_tt,
    "tr": _dash_tr,
    "tb": _dash_tb,
    "tl": _dash_tl,
    "to": _dash_to,
    "rr": _dash_rr,
    "rb": _dash_rb,
    "rl": _dash_rl,
    "ro": _dash_ro,
    "bb": _dash_bb,
    "bl": _dash_bl,
    "bo": _dash_bo,
    "ll": _dash_ll,
    "lo": _dash_lo,
    "oo": _dash_tt,
}



from .mili import (
    MILI,
    MarkDown,
    register_custom_component,
    pack_component,
    get_font_cache,
    clear_font_cache,
    set_number_modifier,
    smart_number,
)
from .data import (
    ElementData,
    Interaction,
    ImageCache,
    TextCache,
    ImageLayerCache,
    ParentCache,
)
from .style import RESIZE, X, CENTER, PADLESS, FLOATING, FILL, SPACELESS
from .utility import (
    Selectable,
    Dragger,
    Scroll,
    Scrollbar,
    Slider,
    DropMenu,
    EntryLine,
    #TextBox,
    ContextMenu,
    GenericApp,
    UIApp,
    AdaptiveUIScaler,
    InteractionSound,
    InteractionCursor,
    CustomWindowBorders,
    CustomWindowBehavior,
    GIF,
    percentage,
    fit_image,
    round_image_borders,
    round_image,
    dict_push,
)
from . import error
from . import style
from . import typing
from . import data
from . import animation
from . import icon

from collections import namedtuple

# explain ContextMenu in guide. Add those XStyleLike to the style guide (uiapp, contextmenu)
VERSION = namedtuple("VersionInfo", "major minor micro")(1, 0, 6)
VERSION_STR = f"{VERSION.major}.{VERSION.minor}.{VERSION.micro}"
del namedtuple

__all__ = (
    "MILI",
    "MarkDown",
    "ElementData",
    "Interaction",
    "ImageCache",
    "TextCache",
    "ImageLayerCache",
    "ParentCache",
    "GIF",
    "Selectable",
    "Dragger",
    "Scroll",
    "Scrollbar",
    "Slider",
    "DropMenu",
    "EntryLine",
    #"TextBox",
    "ContextMenu",
    "GenericApp",
    "UIApp",
    "AdaptiveUIScaler",
    "InteractionSound",
    "InteractionCursor",
    "CustomWindowBorders",
    "CustomWindowBehavior",
    "percentage",
    "fit_image",
    "round_image_borders",
    "round_image",
    "dict_push",
    "register_custom_component",
    "pack_component",
    "get_font_cache",
    "clear_font_cache",
    "set_number_modifier",
    "smart_number",
    "RESIZE",
    "X",
    "CENTER",
    "PADLESS",
    "FLOATING",
    "SPACELESS",
    "FILL",
    "error",
    "style",
    "typing",
    "data",
    "animation",
    "icon",
    "VERSION",
    "VERSION_STR",
)



import sys
import webbrowser


def main():
    sys.stdout.write("\033[F")
    sys.stdout.write("\033[K")
    sys.stdout.flush()

    if len(sys.argv) <= 1:
        print("ERROR: missing command")
        return
    if len(sys.argv) > 2:
        print("ERROR: too many commands")
        return

    command = sys.argv[1]
    help_cmds = ["help", "-h", "--help", "-help", "--h"]
    version_cmds = ["version", "ver", "-v", "--ver", "-version", "--version", "-ver"]

    if (
        command
        not in [
            "guide",
            "changelog",
            "demo",
        ]
        + help_cmds
        + version_cmds
    ):
        print(
            f"ERROR: wrong command '{sys.argv[1]}', expected one of the following: 'guide', 'changelog', 'demo', 'help', 'version'"
        )
        return

    if command == "guide":
        url = "https://github.com/damusss/mili/blob/main/guide/guide.md"
        print(f"Opening guide at '{url}'")
        webbrowser.open(url)
    elif command == "changelog":
        url = "https://github.com/damusss/mili/blob/main/CHANGELOG.md"
        print(f"Opening changelog at '{url}'")
        webbrowser.open(url)
    elif command == "demo":
        from mili.demo import demo

        demo.main()
    elif command in version_cmds:
        import mili

        print(f"MILI {mili.VERSION_STR}")
    elif command in help_cmds:
        print("MILI CLI")
        print("py -m mili guide: open the MILI guide")
        print("py -m mili changelog: open the MILI changelog")
        print("py -m mili version/-v: print the installed MILI version")
        print("py -m mili help/-h: show this message")


if __name__ == "__main__":
    main()



import pygame
import typing
from mili import _core
from mili import data as _data
from mili import typing as _typing
from mili import error as _error
from mili import icon as _icon


class Selectable:
    def __init__(self, selected: bool = False, update_id: str | None = None):
        self.selected: bool = selected
        _core._globalctx._register_update_id(update_id, self.update)

    def update(
        self,
        element: _data.Interaction,
        selectable_group: typing.Sequence["Selectable"] | None = None,
        can_deselect_group: bool = True,
    ) -> _data.Interaction:
        if selectable_group is None:
            selectable_group = []
        if element.left_just_released:
            self.selected = not self.selected
            if len(selectable_group) > 0:
                if self.selected:
                    for sel in selectable_group:
                        if sel is not self:
                            sel.selected = False
                elif not can_deselect_group:
                    one = False
                    for sel in selectable_group:
                        if sel is not self and sel.selected:
                            one = True
                            break
                    if not one:
                        self.selected = True
        elif len(selectable_group) > 0:
            for sel in selectable_group:
                if sel is not self and sel.selected:
                    self.selected = False
                    break
        return element


class Dragger:
    def __init__(
        self,
        position: typing.Sequence[float] | None = None,
        lock_x: bool = False,
        lock_y: bool = False,
        update_id: str | None = None,
    ):
        if position is None:
            position = (0, 0)
        self.position = pygame.Vector2(position)
        self.rel = pygame.Vector2()
        self.lock_x = lock_x
        self.lock_y = lock_y
        self.changed: bool = False
        self.clamp_x: tuple[float, float] | None = None
        self.clamp_y: tuple[float, float] | None = None
        self.style: _typing.ElementStyleLike = {
            "ignore_grid": True,
            "update_id": update_id,
        }
        self._before_update_pos = pygame.Vector2()
        _core._globalctx._register_update_id(update_id, self.update)

    def update(
        self,
        element: _data.Interaction,
    ) -> _data.Interaction:
        if _core._globalctx._mili is None:
            raise _error.MILIStatusError("MILI.start() not called")
        self.changed = False
        self._before_update_pos = self.position.copy()
        if element.left_pressed:
            rel = _core._globalctx._mili._ctx._mouse_rel.copy()
            if self.lock_x:
                rel.x = 0
            if self.lock_y:
                rel.y = 0
            prev = self.position.copy()
            self.position += rel
            if self.position != prev:
                self.changed = True
            self.rel = rel.copy()
        return element

    def clamp(
        self,
        clamp_x: tuple[float, float] | None = None,
        clamp_y: tuple[float, float] | None = None,
        previous_clamps: bool = True,
    ) -> bool:
        if previous_clamps:
            clamp_x = self.clamp_x
            clamp_y = self.clamp_y
        prev = self.position.copy()
        if clamp_x is not None:
            self.position.x = pygame.math.clamp(self.position.x, *clamp_x)
        if clamp_y is not None:
            self.position.y = pygame.math.clamp(self.position.y, *clamp_y)
        if self.position != prev:
            self.changed = True
        if self.changed and self.position == self._before_update_pos:
            self.changed = False
        self.clamp_x = clamp_x
        self.clamp_y = clamp_y
        return self.changed


class Scroll:
    def __init__(self, update_id: str | None = None):
        self.scroll_offset: pygame.Vector2 = pygame.Vector2()
        self._element_data: _data.ElementData | None = None
        self._scrollbars: list["Scrollbar"] = []
        _core._globalctx._register_update_id(update_id, self.update)
        self._update_id = update_id

    def update[IT: _data.Interaction | _data.ElementData](self, container: IT) -> IT:
        if isinstance(container, _data.ElementData):
            self._element_data = container
        else:
            self._element_data = container.data
        self.clamp()
        return container

    def wheel_event(
        self,
        event: pygame.Event,
        multiplier: float = 35,
        constrain_rect: pygame.typing.RectLike | None = None,
        axis_invert_kmod: int | None = None,
    ):
        if event.type == pygame.MOUSEWHEEL and (
            constrain_rect is None
            or pygame.Rect(constrain_rect).collidepoint(pygame.mouse.get_pos())
        ):
            x, y = event.x * multiplier, -event.y * multiplier
            if (
                axis_invert_kmod is not None
                and pygame.key.get_mods() & axis_invert_kmod
            ):
                x, y = y, x
            self.scroll(x, y)

    def scroll(self, x: float, y: float):
        self.scroll_offset.x += x
        self.scroll_offset.y += y
        self.clamp()
        for sbar in self._scrollbars:
            sbar.scroll_moved()

    def set_scroll(self, x: float, y: float):
        self.scroll_offset.x = x
        self.scroll_offset.y = y
        self.clamp()
        for sbar in self._scrollbars:
            sbar.scroll_moved()

    def clamp(self):
        maxx, maxy = float("inf"), float("inf")
        if self._element_data is not None and self._element_data.grid is not None:
            maxx = self._element_data.grid.overflowx
            maxy = self._element_data.grid.overflowy
        self.scroll_offset.x = pygame.math.clamp(self.scroll_offset.x, 0, maxx)
        self.scroll_offset.y = pygame.math.clamp(self.scroll_offset.y, 0, maxy)

    def get_offset(self) -> tuple[float, float]:
        return (-self.scroll_offset.x, -self.scroll_offset.y)

    def get_handle_size(
        self, scrollbar_size: int | float, axis: typing.Literal["x", "y"] = "y"
    ):
        if self._element_data is not None and self._element_data.grid is not None:
            if axis == "x":
                rectsize = self._element_data.rect.w
                overflow = self._element_data.grid.overflowx
            else:
                rectsize = self._element_data.rect.h
                overflow = self._element_data.grid.overflowy
            den = rectsize + overflow
            if den == 0:
                return scrollbar_size
            return min(
                scrollbar_size,
                (rectsize / den) * scrollbar_size,
            )
        return 0

    def get_rel_handle_pos_from_scroll(
        self,
        scrollbar_size: int | float,
        handle_size: int | float,
        axis: typing.Literal["x", "y"] = "y",
    ):
        if self._element_data is None or self._element_data.grid is None:
            return 0
        if axis == "x":
            offset = self.scroll_offset.x
            overflow = self._element_data.grid.overflowx
        else:
            offset = self.scroll_offset.y
            overflow = self._element_data.grid.overflowy
        if overflow == 0:
            return 0
        return (offset / overflow) * (scrollbar_size - handle_size)

    def set_scroll_from_rel_handle_pos(
        self,
        scrollbar_size: int | float,
        rel_handle_pos: int | float,
        handle_size: int,
        axis: typing.Literal["x", "y"] = "y",
    ):
        if self._element_data is None or self._element_data.grid is None:
            return
        if scrollbar_size - handle_size == 0:
            return
        if axis == "x":
            self.scroll_offset.x = self._element_data.grid.overflowx * (
                rel_handle_pos / (scrollbar_size - handle_size)
            )
        else:
            self.scroll_offset.y = self._element_data.grid.overflowy * (
                rel_handle_pos / (scrollbar_size - handle_size)
            )
        self.clamp()


class Scrollbar:
    def __init__(
        self,
        scroll: Scroll,
        style: _typing.ScrollbarStyleLike | dict | None = None,
    ):
        if style is None:
            style = {}
        self._scroll = scroll
        self._scroll._scrollbars.append(self)
        self.style: _typing.ScrollbarStyleLike = {
            "axis": style.get("axis", "y"),
            "short_size": style.get("short_size", 10),
            "border_dist": style.get("border_dist", 3),
            "padding": style.get("padding", 3),
            "size_reduce": style.get("size_reduce", 0),
            "handle_update_id": style.get("handle_update_id", None),
        }
        self.size: int | float = 0
        self.handle_size: int = 0
        self.handle_dragger: Dragger = Dragger(
            lock_x=self.axis == "y", lock_y=self.axis == "x"
        )
        self.bar_rect: pygame.Rect = pygame.Rect()
        self.handle_rect: pygame.Rect = pygame.Rect()

        self.bar_style: _typing.ElementStyleLike = {"ignore_grid": True, "z": 999}
        self.handle_style: _typing.ElementStyleLike = {
            "ignore_grid": True,
            "z": 9999,
            "update_id": self.style["handle_update_id"],
        }

        _core._globalctx._register_update_id(self._scroll._update_id, self.update)
        _core._globalctx._register_update_id(
            self.style["handle_update_id"], self.update_handle
        )

    @property
    def axis(self) -> typing.Literal["x", "y"]:
        return self.style["axis"]

    @axis.setter
    def axis(self, value):
        self.axis = value
        self.handle_dragger.lock_x = self.axis == "x"
        self.handle_dragger.lock_y = self.axis == "y"

    @property
    def needed(self):
        if (
            self._scroll._element_data is None
            or self._scroll._element_data.grid is None
        ):
            return False
        return (
            self._scroll._element_data.grid.overflowy
            if self.axis == "y"
            else self._scroll._element_data.grid.overflowx
        ) > 0

    def update[IT: _data.Interaction | _data.ElementData](self, container: IT) -> IT:
        if isinstance(container, _data.ElementData):
            element_data = container
        else:
            element_data = container.data
        rectsize = element_data.rect.h if self.axis == "x" else element_data.rect.w
        previous = self.size
        self.size = (
            (element_data.rect.w if self.axis == "x" else element_data.rect.h)
            - self.style["padding"] * 2
            - self.style["size_reduce"]
        )
        if previous != self.size and previous != 0:
            inc = self.size / previous
            if self.axis == "x":
                self.handle_dragger.position.x *= inc
            else:
                self.handle_dragger.position.y *= inc
            self.handle_dragger.changed = True
        if self.axis == "x":
            self.bar_rect = pygame.Rect(
                self.style["padding"],
                rectsize - self.style["border_dist"] - self.style["short_size"],
                self.size,
                self.style["short_size"],
            )
        else:
            self.bar_rect = pygame.Rect(
                rectsize - self.style["border_dist"] - self.style["short_size"],
                self.style["padding"],
                self.style["short_size"],
                self.size,
            )
        self.handle_size = int(self._scroll.get_handle_size(self.size, self.axis))
        return container

    def update_handle(self, handle_element: _data.Interaction) -> _data.Interaction:
        changed = self.handle_dragger.changed
        self.handle_dragger.update(
            handle_element,
        )
        self.handle_dragger.clamp(
            clamp_x=(0, self.size - self.handle_size) if self.axis == "x" else None,
            clamp_y=(0, self.size - self.handle_size) if self.axis == "y" else None,
            previous_clamps=False,
        )
        if self.handle_dragger.changed or changed:
            self._scroll.set_scroll_from_rel_handle_pos(
                self.size,
                int(
                    self.handle_dragger.position.x
                    if self.axis == "x"
                    else self.handle_dragger.position.y
                ),
                self.handle_size,
                self.axis,
            )
        if self.axis == "x":
            self.handle_rect = pygame.Rect(
                self.handle_dragger.position.x,
                0,
                self.handle_size,
                self.style["short_size"],
            )
        else:
            self.handle_rect = pygame.Rect(
                0,
                self.handle_dragger.position.y,
                self.style["short_size"],
                self.handle_size,
            )
        return handle_element

    def scroll_moved(self):
        pos = self._scroll.get_rel_handle_pos_from_scroll(
            self.size, self.handle_size, self.axis
        )
        if self.axis == "x":
            self.handle_dragger.position.x = pos
        else:
            self.handle_dragger.position.y = pos
        self.handle_dragger.clamp()


class Slider:
    def __init__(
        self,
        style: _typing.SliderStyleLike | dict | None = None,
    ):
        if style is None:
            style = {}
        self.style: _typing.SliderStyleLike = {
            "lock_x": style.get("lock_x", False),
            "lock_y": style.get("lock_y", False),
            "handle_size": style.get("handle_size", (10, 10)),
            "strict_borders": style.get("strict_borders", False),
            "area_update_id": style.get("area_update_id", None),
            "handle_update_id": style.get("handle_update_id", None),
            "drag_area": style.get("drag_area", True),
        }
        self._area_data: _data.ElementData | None = None
        self._start_value: pygame.Vector2 | None = None
        self.dragging_area = False
        self.handle_dragger: Dragger = Dragger()
        self.moved: bool = False

        self.area_style: _typing.ElementStyleLike = {
            "clip_draw": False,
            "update_id": self.style["area_update_id"],
        }
        self.handle_style: _typing.ElementStyleLike = {
            "ignore_grid": True,
            "update_id": self.style["handle_update_id"],
        }
        self.handle_rect: pygame.Rect = pygame.Rect()

        _core._globalctx._register_update_id(
            self.style["area_update_id"], self.update_area
        )
        _core._globalctx._register_update_id(
            self.style["handle_update_id"], self.update_handle
        )

    @classmethod
    def from_axis(
        cls,
        axis: typing.Literal["x", "y"],
        style: _typing.SliderStyleLike | dict | None,
    ):
        if style is None:
            style = {}
        return cls({**style, "lock_x": axis == "y", "lock_y": axis == "x"})

    def update_area(self, area_element: _data.Interaction) -> _data.Interaction:
        previous = self.value
        self._area_data = area_element.data
        if (
            self._area_data is not None
            and self._start_value is not None
            and (self._area_data.rect.w != 0 or self._area_data.rect.h != 0)
        ):
            self.value = self._start_value
            self._start_value = None
        else:
            if self.value != previous:
                self.value = previous
        if self.style["drag_area"] and area_element.left_pressed:
            before = self.handle_dragger.position.copy()
            diff = (
                pygame.Vector2(pygame.mouse.get_pos())
                - self._area_data.absolute_rect.topleft
            )
            self.dragging_area = True
            clamp_padx, clamp_pady = self._pre_update()
            self.handle_dragger.position = diff
            self._post_update(clamp_padx, clamp_pady)
            self.moved = before != self.handle_dragger.position
        else:
            self.dragging_area = False
        return area_element

    def update_handle(self, handle_element: _data.Interaction) -> _data.Interaction:
        if self.dragging_area:
            return handle_element
        if self._area_data is None:
            raise _error.MILIStatusError(
                "Slider.update_area must be called before Slider.update_handle"
            )
        clamp_padx, clamp_pady = self._pre_update()
        self.handle_dragger.update(
            handle_element,
        )
        self._post_update(clamp_padx, clamp_pady)
        self.moved = self.handle_dragger.changed
        return handle_element

    def _post_update(self, clamp_padx, clamp_pady):
        if self._area_data is None:
            raise
        self.handle_dragger.clamp(
            (0 + clamp_padx, self._area_data.rect.w - clamp_padx),
            (0 + clamp_pady, self._area_data.rect.h - clamp_pady),
            False,
        )
        if self.style["lock_x"]:
            self.handle_dragger.position.x = self._area_data.rect.w / 2
        if self.style["lock_y"]:
            self.handle_dragger.position.y = self._area_data.rect.h / 2
        self.handle_rect = pygame.Rect(
            (
                self.handle_dragger.position.x - self.style["handle_size"][0] / 2,
                self.handle_dragger.position.y - self.style["handle_size"][1] / 2,
            ),
            self.style["handle_size"],
        )

    def _pre_update(self):
        if self._area_data is None:
            raise
        self.handle_dragger.lock_x = self.style["lock_x"]
        self.handle_dragger.lock_y = self.style["lock_y"]
        clamp_padx, clamp_pady = (
            self.style["handle_size"][0] / 2
            if self.style["strict_borders"]
            and self._area_data.rect.w >= self.style["handle_size"][0]
            else 0,
            self.style["handle_size"][1] / 2
            if self.style["strict_borders"]
            and self._area_data.rect.h >= self.style["handle_size"][1]
            else 0,
        )
        return clamp_padx, clamp_pady

    @property
    def value(self) -> pygame.Vector2:
        if self._area_data is None:
            if self._start_value is not None:
                return self._start_value
            return pygame.Vector2(-1, -1)
        if not self.style["strict_borders"]:
            return pygame.Vector2(
                0
                if self.style["lock_x"] or self._area_data.rect.w == 0
                else self.handle_dragger.position.x / self._area_data.rect.w,
                0
                if self.style["lock_y"] or self._area_data.rect.h == 0
                else self.handle_dragger.position.y / self._area_data.rect.h,
            )
        hx, hy = self.style["handle_size"][0] / 2, self.style["handle_size"][1] / 2
        return pygame.Vector2(
            0
            if self.style["lock_x"] or self._area_data.rect.w == 0
            else (self.handle_dragger.position.x - hx)
            / (self._area_data.rect.w - hx * 2),
            0
            if self.style["lock_y"] or self._area_data.rect.h == 0
            else (self.handle_dragger.position.y - hy)
            / (self._area_data.rect.h - hy * 2),
        )

    @value.setter
    def value(self, v: typing.Sequence[float] | pygame.Vector2):
        v = (pygame.math.clamp(v[0], 0, 1), pygame.math.clamp(v[1], 0, 1))
        if self._area_data is None:
            self._start_value = pygame.Vector2(v)
            return
        if not self.style["strict_borders"]:
            if not self.style["lock_x"]:
                self.handle_dragger.position.x = v[0] * self._area_data.rect.w
            if not self.style["lock_y"]:
                self.handle_dragger.position.y = v[1] * self._area_data.rect.h
            self.handle_rect = pygame.Rect(
                (
                    self.handle_dragger.position.x - self.style["handle_size"][0] / 2,
                    self.handle_dragger.position.y - self.style["handle_size"][1] / 2,
                ),
                self.style["handle_size"],
            )
            return
        hx, hy = self.style["handle_size"][0] / 2, self.style["handle_size"][1] / 2
        if not self.style["lock_x"]:
            self.handle_dragger.position.x = (
                v[0] * (self._area_data.rect.w - hx * 2) + hx
            )
        if not self.style["lock_y"]:
            self.handle_dragger.position.y = (
                v[1] * (self._area_data.rect.h - hy * 2) + hy
            )
        self.handle_rect = pygame.Rect(
            (
                self.handle_dragger.position.x - self.style["handle_size"][0] / 2,
                self.handle_dragger.position.y - self.style["handle_size"][1] / 2,
            ),
            self.style["handle_size"],
        )

    @property
    def valuex(self) -> float:
        return self.value.x

    @property
    def valuey(self) -> float:
        return self.value.y

    @valuex.setter
    def valuex(self, v: float):
        self.value = pygame.Vector2(v, self.value.y)

    @valuey.setter
    def valuey(self, v: float):
        self.value = pygame.Vector2(self.value.x, v)


class DropMenu:
    def __init__(
        self,
        options: typing.Sequence[typing.Any],
        selected: typing.Any,
        style: _typing.DropMenuStyleLike | dict | None = None,
        ui_style: _typing.DropMenuUIStyleLike | dict | None = None,
    ):
        if style is None:
            style = {}
        if ui_style is None:
            ui_style = {}
        self.options = list(options)
        self.selected = selected
        self.just_selected: bool = False
        self.shown: bool = False
        self.style: _typing.DropMenuStyleLike = {
            "direction": style.get("direction", "down"),
            "anchor": style.get("anchor", "first"),
            "padding": style.get("padding", 3),
            "selected_update_id": style.get("selected_update_id", None),
            "option_update_id": style.get("option_update_id", None),
            "menu_update_id": style.get("menu_update_id", None),
        }
        opt_cols = ui_style.get("option_colors", {})
        selopt_cols = ui_style.get("selected_option_colors", {})
        self.ui_style: _typing.DropMenuUIStyleLike = {
            "bg_color": ui_style.get("bg_color", (24, 24, 24)),
            "border_radius": ui_style.get("border_radius", 7),
            "option_border_radius": ui_style.get("option_border_radius", 7),
            "option_color": {
                "default": opt_cols.get("default", None),
                "hover": opt_cols.get("hover", (40, 40, 40)),
                "press": opt_cols.get("press", (40, 40, 40)),
            },
            "selected_option_color": {
                "default": selopt_cols.get("default", (24, 24, 24)),
                "hover": selopt_cols.get("hover", (40, 40, 40)),
                "press": selopt_cols.get("press", (40, 40, 40)),
            },
            "outline_color": ui_style.get("outline_color", (40, 40, 40)),
            "menu_style": ui_style.get("menu_style", {"pad": 3, "spacing": 0}),
            "text_style": ui_style.get("text_style", {"align": "left"}),
            "icon_arrow_down": ui_style.get("icon_arrow_down", "arrow_drop_down"),
            "icon_arrow_up": ui_style.get("icon_arrow_up", "arrow_drop_up"),
            "option_text_style": ui_style.get("option_text_style", {"align": "left"}),
        }
        self.topleft = pygame.Vector2(0, 0)
        self.menu_style: _typing.ElementStyleLike = {
            "ignore_grid": True,
            "update_id": self.style["menu_update_id"],
        }
        self._menu_rect: pygame.Rect | None = None
        self._menu_parent_abspos = pygame.Vector2()
        self._option_i = 0
        self.__just_open = False
        _core._globalctx._register_update_id(
            self.style["selected_update_id"], self.update_selected
        )
        _core._globalctx._register_update_id(
            self.style["option_update_id"], self.update_option
        )
        _core._globalctx._register_update_id(
            self.style["menu_update_id"], self.update_menu
        )

    def show(self):
        self.shown = True

    def hide(self):
        self.shown = False

    def toggle(self):
        self.shown = not self.shown

    def update_selected(self, selected_element: _data.Interaction) -> _data.Interaction:
        if selected_element.left_just_released:
            self.toggle()
            self.__just_open = True
        else:
            self.__just_open = False
        if self._menu_rect is None or self._menu_rect.h == 0:
            return selected_element
        menur = self._menu_rect
        selr = selected_element.data.absolute_rect
        match self.style["anchor"]:
            case "center":
                self.topleft = pygame.Vector2(
                    selr.x - (menur.w - selr.w) / 2 - self._menu_parent_abspos.x, 0
                )
            case "first":
                self.topleft = pygame.Vector2(selr.x - self._menu_parent_abspos.x, 0)
            case "last":
                self.topleft = pygame.Vector2(
                    selr.x - (menur.w - selr.w) - self._menu_parent_abspos.x, 0
                )
        if self.style["direction"] == "down":
            self.topleft.y = (
                selr.bottom + self.style["padding"] - self._menu_parent_abspos.y
            )
        else:
            self.topleft.y = (
                selr.top - menur.h - self.style["padding"] - self._menu_parent_abspos.y
            )
        if not self.shown and _core._globalctx._mili is not None:
            parent = _core._globalctx._mili.data_from_id(
                selected_element.data.parent_id
            )
            if parent is not None:
                self._menu_parent_abspos = pygame.Vector2(parent.absolute_rect.topleft)
        return selected_element

    def update_menu(self, menu_element: _data.Interaction) -> _data.Interaction:
        self._option_i = 0
        if menu_element.data.rect.h != 0:
            self._menu_rect = menu_element.data.rect
        if _core._globalctx._mili is None:
            self._menu_parent_abspos = pygame.Vector2(0, 0)
            return menu_element
        parent = _core._globalctx._mili.data_from_id(menu_element.data.parent_id)
        if parent is None:
            self._menu_parent_abspos = pygame.Vector2(0, 0)
            return menu_element
        self._menu_parent_abspos = pygame.Vector2(parent.absolute_rect.topleft)
        return menu_element

    def update_option(self, option_element: _data.Interaction) -> _data.Interaction:
        self.just_selected = False
        if option_element.left_just_released and not self.__just_open:
            self.selected = self.options[self._option_i]
            self.just_selected = True
            self.hide()
        self._option_i += 1
        return option_element

    def ui(
        self,
        selected_rect: pygame.typing.RectLike | None,
        selected_style: _typing.ElementStyleLike | None,
    ):
        upicon, downicon = (
            _icon._get_icon(self.ui_style["icon_arrow_up"]),
            _icon._get_icon(self.ui_style["icon_arrow_down"]),
        )
        _mili = _core._globalctx._mili
        if _mili is None:
            raise _error.MILIStatusError(
                "A MILI instance must be started to use this utility method"
            )
        with _mili.begin(
            selected_rect,
            {"pad": 0, "spacing": 0, "update_id": "cursor"}
            | (selected_style if selected_style else {})
            | {"axis": "x"},
        ) as selected:
            _mili.rect(
                {
                    "color": _core._coreutils._get_conditional_color(
                        selected, self.ui_style["selected_option_color"]
                    ),
                    "border_radius": self.ui_style["border_radius"],
                }
            )
            _mili.rect(
                {
                    "color": self.ui_style["outline_color"],
                    "border_radius": self.ui_style["border_radius"],
                    "outline": 1,
                    "draw_above": True,
                }
            )
            _mili.text_element(
                self.selected,
                {"growx": False} | self.ui_style["text_style"],
                None,
                {"fillx": True, "filly": True, "blocking": False},
            )
            surf = (
                (upicon if self.shown else downicon)
                if self.style["direction"] == "down"
                else (downicon if self.shown else upicon)
            )
            if surf:
                height = selected.data.rect.h
                ratio = surf.width / surf.height
                width = height * ratio
                _mili.image_element(
                    surf, None, (0, 0, width, height), {"blocking": False}
                )
            self.update_selected(selected)
            sel_width = selected.data.rect.w
        if self.shown:
            with _mili.begin(
                (self.topleft, (sel_width, 0)),
                {"z": 99999}
                | self.ui_style["menu_style"]
                | {"resizey": True, "ignore_grid": True},
            ) as menuel:
                _mili.rect(
                    {
                        "color": self.ui_style["bg_color"],
                        "border_radius": self.ui_style["border_radius"],
                    }
                )
                _mili.rect(
                    {
                        "color": self.ui_style["outline_color"],
                        "border_radius": self.ui_style["border_radius"],
                        "outline": 1,
                        "draw_above": True,
                    }
                )
                self.update_menu(menuel)
                for i, option in enumerate(self.options):
                    br = self.ui_style["option_border_radius"]
                    if i == 0:
                        br = (
                            self.ui_style["border_radius"],
                            self.ui_style["border_radius"],
                            br,
                            br,
                        )
                    elif i == len(self.options) - 1:
                        br = (
                            br,
                            br,
                            self.ui_style["border_radius"],
                            self.ui_style["border_radius"],
                        )
                    with _mili.element(
                        None, {"fillx": True, "update_id": "cursor"}
                    ) as optel:
                        _mili.rect(
                            {
                                "color": _core._coreutils._get_conditional_color(
                                    optel, self.ui_style["option_color"]
                                ),
                                "border_radius": br,
                            }
                        )
                        _mili.text(
                            option, {"growy": True} | self.ui_style["option_text_style"]
                        )
                        self.update_option(optel)
        else:
            with _mili.begin(
                (self.topleft, (sel_width, 0)),
                {"z": 99999}
                | self.ui_style["menu_style"]
                | {"resizey": True, "ignore_grid": True, "pad": 0, "spacing": 0},
            ) as menuel:
                self.update_menu(menuel)
                for opt in self.options:
                    _mili.element(None)
